import gensim
from compling.embeddings.words.wembeddings import WordEmbeddings
from typing import Iterable


class Fasttext(WordEmbeddings):
    """
    FastText is able to achieve really good performance for word representations and sentence classification, specially in the case of rare words by making use of character level information.

    Each word is represented as a bag of character n-grams in addition to the word itself. This helps preserve the meaning of shorter words that may show up as ngrams of other words. Inherently, this also allows you to capture meaning for suffixes/prefixes.
    """
    def __init__(self, index: Iterable[dict], output:str=None, text_field='text', skipgram_ws: int = None):
        super().__init__(output=output, index=index, text_field=text_field, tag=False, skipgram_ws=skipgram_ws,
                         module=gensim.models.fasttext.FastText)


import pandas as pd
import matplotlib.cm as cm
import numpy as np
import matplotlib.pyplot as plt
from sklearn import decomposition
from compling.unsupervisedLearning.unsupervised_learning import UnsupervisedLearning
from typing import Tuple


class TruncatedSVD(UnsupervisedLearning):
    """
    TruncateSVD, is a dimensionality-reduction method that is often used to reduce the dimensionality of large data sets, by transforming a large set of variables into a smaller one that still contains most of the information in the large set.
    """

    def run(self, n:int, **kwargs) -> object:
        """
        Runs TruncateSVD on input vectors.

        Args:
            n (int): Number of components to keep.
            kwargs (object): See https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.TruncateSVD.html .

        Returns:
            object: Performed dimensionality reduction on input vectors. <br/> See https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.TruncateSVD.html
        """

        truncatesvd = decomposition.TruncatedSVD(n_components=n)
        self.truncatesvd_data = pd.DataFrame(truncatesvd.fit_transform(self.data.T), index=self.data.T.index,
                                             columns=['X', 'Y'])
        return self.truncatesvd_data

    def plot(self, legend: bool = True, figsize: Tuple[int, int] = (10, 8),
             title: str = '', xlim: Tuple[float, float] = None, ylim: Tuple[float, float] = None,
             hidexticks: bool = False, save: bool = None) -> None:
        """
        Plots the reduced data (if 2D).

        Args:
            legend (bool, optional, default=True): If True, plots legend on the figure.
            figsize (Tuple[int, int], optional, default=10,8): Figure size: width, height in inches.
            title (str, optional, default=None): If not None, draws title string on the figure.
            xlim (Tuple[int, int], optional, default=None): Limits the values on the x-axis to the range (x_min, x_max).
            ylim (Tuple[int, int], optional, default=None): Limits the values on the y-axis to the range (y_min, y_max).
            hidexticks (bool, optional, default=False): If True, hides xticks.
            save (str, optional, default=None): Figure is not saved if save is None. <br/>  Set it with a output name if you want save it.
        """

        if self.truncatesvd_data.shape[1] != 2:
            raise Exception("data has too components. You can plot only bidimensional data.")

        colors = iter(cm.rainbow(np.linspace(0, 1, self.truncatesvd_data.shape[0])))
        fig, ax = plt.subplots(figsize=figsize)
        for i, u in enumerate(self.truncatesvd_data.index):
            ax.scatter(self.truncatesvd_data.loc[u].X, self.truncatesvd_data.loc[u].Y,
                       label=u, s=100, alpha=0.5, c=np.array([next(colors)]))

        # Put a legend to the right of the current axis
        if legend:
            ax.legend(loc='center left', bbox_to_anchor=(1, 0.5))

        if xlim is not None:
            plt.xlim(xlim[0], xlim[1])
        if ylim is not None:
            plt.ylim(ylim[0], ylim[1])

        if not hidexticks:
            ax.set_xticks([])
            ax.set_yticks([])

        if not title is None:
            plt.title(title)

        if save:
            fig1 = plt.gcf()
            plt.show()
            plt.draw()
            fig1.savefig(save)
        else:
            plt.show()

        plt.close()



language = 'spanish'
iso639 = 'es'

spacy_model_sm = iso639 + '_core_news_sm'
spacy_model_md = iso639 + '_core_news_md'
spacy_model_lg = iso639 + '_core_news_lg'

sentiment_lexicon_class = None
pos_token_sentiment = None

negations = ['nadie', 'nada', 'ni', 'jamás', 'jamas', 'nunca', 'tampoco', 'todavía no', 'todavia no',
                    'ya no', 'no', 'ninguno', 'ningún', 'ningun', 'ni siquiera', 'más']


import csv
import pkg_resources
from compling.analysis.sentiment.lexicon.lexicon import Lexicon
from collections import defaultdict
from typing import Tuple

csv.field_size_limit(1000000)

__sentix__ = pkg_resources.resource_filename('compling', 'senti-lexicons/sentix.csv')

class Sentix(Lexicon):
    """Sentix Lexicon Implementation of Lexicon Abastract Class.

    Italian Language.

    http://valeriobasile.github.io/twita/sentix.html
    """

    def __init__(self, pos: Tuple[str, str] = None) -> None:
        """
        **\_\_init\_\_**: Creates a new Sentiwordnet Italian object.

        Args:
            pos (Tuple[str], optional, default=None): SentiWordNet Lexicon returns a token polarities for each part of speech that token may have in the speech. <br/> If not None, filters the token polarities for part of speech in the pos tuple.

        This class sums the polarity value of the parts of speech you choose
        """

        self.sentiment = defaultdict(lambda: defaultdict(list))

        with open(__sentix__, encoding='utf-8', mode='r') as f:
            rows = csv.reader(f, delimiter='\t')

            columns = next(rows)

            for row in rows:
                record = dict(zip(columns, row))
                lemma, synset = record['lemma'].lower(), record['POS']
                self.sentiment[lemma][synset].append(record)

        pos_dict = {'n': 'NOUN', 'v': 'VERB', 'a': 'ADJ', 's': 'ADJ', 'r': 'ADV'}
        self.pos = [k for k, v in pos_dict.items() if pos is None or v in pos or len(pos) == 0]

    def polarity(self, token: str) -> dict:
        """
        Returns the token polarities.

        Args:
            token (str): The text of the input token.

        Returns:
            dict: Polarities type as keys, sentiment values as values.

                    Example:
                    {
                     'pos' : 0.9,
                     'neg' : 0.1,
                     'obj' : 0.3
                    }
        """

        polarities = dict({'neg': 0.0, 'pos': 0.0, 'obj': 0.0})
        count = dict({'neg': 0, 'pos': 0, 'obj': 0})

        if token.replace("NOT_", "").lower() not in self.sentiment:
            return polarities

        for synset in self.sentiment[token.replace("NOT_", "").lower()]:
            if not synset in self.pos:
                continue

            for record in self.sentiment[token.replace("NOT_", "").lower()][synset]:

                polarity = float(record['polarity'])
                pos = 0 if polarity < 0 else polarity
                neg = 0 if polarity > 0 else -polarity
                obj = 1 - float(record['intensity'])


                if token.startswith('NOT_'):

                    if neg > 0:
                        count['pos'] += 1
                    if pos > 0:
                        count['neg'] += 1

                    polarities['neg'] += pos
                    polarities['pos'] += neg
                else:

                    if pos > 0:
                        count['pos'] += 1
                    if neg > 0:
                        count['neg'] += 1

                    polarities['pos'] += pos
                    polarities['neg'] += neg

                if obj > 0:
                    count['obj'] +=1

                polarities['obj'] += obj

        for k in polarities.keys():
            if count[k] > 0:
                if k in ("pos", "neg"):
                    polarities[k] /= max(count["pos"], count["neg"])
                else:
                    polarities[k] /= count[k]

        return polarities

if __name__ == '__main__':
    s = Sentix()
    print(s.polarity('essere'))


language = 'chinese'
iso639 = 'zh'

spacy_model_sm = iso639 + '_core_news_sm'
spacy_model_md = iso639 + '_core_news_md'
spacy_model_lg = iso639 + '_core_news_lg'

sentiment_lexicon_class = 'Fannix'
pos_token_sentiment = ['AUX', 'VERB', 'ADJ', 'ADV', 'NOUN', 'PROPN']

negations = ['bù', '不', 'méi', '没', 'méiyǒu', '没有', "甚至", "完全不", "少", "決不", "不再"]



language = 'portuguese'
iso639 = 'pt'

spacy_model_sm = iso639 + '_core_news_sm'
spacy_model_md = iso639 + '_core_news_md'
spacy_model_lg = iso639 + '_core_news_lg'

sentiment_lexicon_class = None
pos_token_sentiment = None

negations = ['nada', 'ninguém', 'ninguem', 'nenhum', 'tampouco', 'nem', 'nunca', 'jamais', 'não', 'nao',
                    'nenhum', 'nenhuma', 'nem mesmo']


from nltk.corpus import sentiwordnet as swn
from compling.lexicons.lexiconMeta import LexiconMeta


class Sentiwordnet(LexiconMeta):
    """Sentiwordnet Lexicon for English Language."""

    def __init__(self):
        module = __import__('compling.languages.english')
        language = getattr(getattr(module, 'languages'), 'english')
        super().__init__(language.pos_token_sentiment)

    def polarity(self, token: str) -> dict:
        """
        Returns a float for sentiment strength based on the input text.

        * The positive polarity of a token is calculated as the average of the positive polarities of its synsets that have a positive or negative polarity greater than 0.
        * The negative polarity of a token is calculated as the average of the positive polarities of its synsets that have a positive polarity or negative greater than 0.
        * The neutral polarity of a token is calculated as the average of the positive polarities of its synsets that have a neutral polarity greater than 0.

        Args:
            token (str): The text of the input token.

        Returns:
            dict: A positive, a negative and a neutral value.
        """

        # text processing
        is_negated = True if token.startswith('NOT_') else False
        if is_negated:
            token = token.replace("NOT_", "")
        token = token.lower()

        senti_synsets = [senti_synset for senti_synset in
                         list(swn.senti_synsets(token))
                         if senti_synset.synset._pos in self.__PoS__]

        polarities = dict({'neg': 0.0, 'pos': 0.0, 'obj': 0.0})
        count = dict({'neg': 0, 'pos': 0, 'obj': 0})

        for senti_synset in senti_synsets:
            pos = senti_synset.pos_score()
            neg = senti_synset.neg_score()
            obj = senti_synset.obj_score()

            # Inversion of polarity
            if is_negated:
                pos, neg = neg, pos

            if pos > 0:
                count['pos'] += 1
            if neg > 0:
                count['neg'] += 1

            polarities['pos'] += pos
            polarities['neg'] += neg

            if obj > 0:
                count['obj'] += 1
            polarities['obj'] += obj

        # Normalization
        if count['pos'] > 0 or count['neg'] > 0:
            polarities['pos'] /= max(count["pos"], count["neg"])
            polarities['neg'] /= max(count["pos"], count["neg"])
        if count['obj'] > 0:
            polarities['obj'] /= count['obj']

        return polarities



import gensim
from compling.embeddings.words.wembeddings import WordEmbeddings
from typing import Iterable


class Word2vec(WordEmbeddings):
    """
    Word2Vec is one of the most popular technique to learn word embeddings using shallow neural network. It was developed by Tomas Mikolov in 2013 at Google.

    It can be obtained using two methods (both involving Neural Networks): Skip Gram and Common Bag Of Words (CBOW)
    CBOW Model: This method takes the context of each word as the input and tries to predict the word corresponding to the context.

    Skip – gram follows the same topology as of CBOW. It just flips CBOW’s architecture on its head. The aim of skip-gram is to predict the context given a word.
    """
    def __init__(self, index: Iterable[dict], output:str=None, text_field='text', skipgram_ws: int = None):
        super().__init__(output=output, index=index, text_field=text_field, tag=False,
                         skipgram_ws=skipgram_ws, module=gensim.models.Word2Vec)



import abc
from typing import *


class SentimentLexiconBased(metaclass=abc.ABCMeta):
    """
    The most popular unsupervised strategies used in _sentiment analysis_ are _lexical-based_ methods. <br/>
    They make use of a predefined list of words, where each word is associated with a specific _sentiment_. <br/>
    Lexicon-based strategies are very efficient and simple methods. They make use of a sentiment _lexicon_
     to assign a polarity value to each text document by following a basic algrithm. A sentiment lexicon is a
     list of lexical features (e.g., words, phrase, etc.) which are labeled according to their semantic
     orientation (i.e. polarity) as either _positive_ or _negative_.
    """

    def __init__(self, group_by_field: Union[str, List[str]], date_field: str = None) -> None:
        """
        **\_\_init\_\_**: Creates a new SentimentAnalyzer object.

        Args:
            group_by_field (Union[str, List[str]]): Sentiments values are calculated for each different tuple of values. Tuples are defined by grouping records by fields in group_by_field.
            date_field (str, optional, default=None): If not None, Sentiment vectors are sorted by this field.

        """
        if isinstance(group_by_field, str):
            self.group_by_field = [group_by_field + '_']
        else:
            self.group_by_field = [g + '_' for g in group_by_field]

        self.group_by_field_key = " + ".join(self.group_by_field)

        self.date_field = date_field if date_field is None else date_field + '_'

    @abc.abstractmethod
    def filter(self, index: Iterable[dict],
               regex_pattern: str,
               text_field: str = 'text') -> Iterable[dict]:
        """
        Filters the tokens/sentences/paragraphs/documents to be considered to calculate a sentiment value.

        Args:
            index (Iterable[dict]): Records of an index produced during the tokenization stage (e.g. sentences, paragraphs, documents).
            regex_pattern (str): The filter uses this regex.
            text_field (str, optional, default=text): For each record the regex will be evaluated on this field.
        Returns:
            Iterable[dict]: The record filtered.
        """

    @abc.abstractmethod
    def run(self, index: Iterable[dict],
            text_field: str = 'text',
            min_len: int = 0,
            lang: str = None,
            pos: Tuple[str, ...] = None,
            dep: Tuple[str, ...] = None) -> Dict[str, Dict[str, float]]:
        """
        Performs sentiment analysis of the texts in the index calculating the polarity level for each different
        group_by_field value.

        Args:
            index (Iterable[dict]): Records of an index produced during the tokenization stage (e.g. sentences, paragraphs, documents).

                    Example:
                    >> s = SentimentAnalyzer(group_by_field='author')
                    >> sentences_filtered = s.filtered(sentences, regex_pattern='.*family.*')
                    >> s.run(index=sentences_filtered)

            text_field (str, optional, default=text): For each record 'min_len' will be checked on this field.
            min_len (int, optional, default=0): The minimum length a text must have to be considered
            lang (str, optional, default=None): You can filter the sentiment analysis for the document of a specific language.
            pos (Tuple[str, ...], optional, default=None): If index records are tokens, part of speech of the tokens. <br/> Filters the most informative tokens: tokens whose PoS tag is in the 'pos' list. <br/> If None, the filter will be ignored.
            dep (Tuple[str, ...], optional, default=None): If index records are tokens, dependencies of the tokens. <br/> Filters the most informative tokens: tokens whose dep tag is in the 'dep' list. <br/> If None, the filter will be ignored.

        Returns:
            Dict[str, Dict[str, float]]: A dict contains the polarity value for each different grouped_by_field value.

                         Example:
                         {
                          "grouped_by_field1": {
                                                "pos": 0,
                                                "neg":1,
                                                "obj":3},
                           ...,
                         }
        """

    @abc.abstractmethod
    def sentitokens(self,
                    tokens: Iterable[dict],
                    text_field: str = 'text',
                    min_len: int = 0,
                    lang: str = None,
                    pos: Tuple[str, ...] = None,
                    dep: Tuple[str, ...] = None) -> dict:
        """

        Args:
            tokens (Iterable[dict]): Stream of token records produced during the Tokenization stage.
            text_field (str, optional, default=text): The text field of each token.
            min_len (int, optional, default=0): The minimum length a token must have to be considered
            lang (str, optional, default=None): You can filter the token of a specific language.
            pos (Tuple[str, ...], optional, default=None): Part of speech of the tokens. <br/> Filters the most informative tokens: tokens whose PoS tag is in the 'pos' list. <br/> If None, the filter will be ignored.
            dep (Tuple[str, ...], optional, default=None): Dependencies of the tokens. <br/> Filters the most informative tokens: tokens whose dep tag is in the 'dep' list. <br/> If None, the filter will be ignored.

        Returns:
            dict: A dict containing the words classified as positive/negative, as keys, with their frequency value, as values, for each
            different grouped_by_field value.

                         Example:
                         {
                          "grouped_by_field1": {
                                                "neg": {
                                                        "fuck": 12,
                                                        "bad": 7,
                                                        ...
                                                        },
                                                "pos": {
                                                        "good": 11,
                                                        "love": 2,
                                                        ...
                                                        }
                                                },
                          ...
                        }
        """


language = 'english'
iso639 = 'en'

spacy_model_sm = iso639 + '_core_web_sm'
spacy_model_md = iso639 + '_core_web_md'
spacy_model_lg = iso639 + '_core_web_lg'

sentiment_lexicon_class = 'Sentiwordnet'

# Some lexicons provide a polarity of a token for each part of speech that token may have in the speech.
# If not None, only the polarities for the parts of speech in 'pos_token_sentiment' will be considered.
pos_token_sentiment = ['AUX', 'VERB', 'ADJ', 'ADV', 'NOUN', 'PROPN']

negations = ['not', "n't", 'no', 'none', 'neither', 'nor', 'never', 'hardly', 'scarcely',
             'barely', 'rarely', 'seldom', "doesn't", "isn't", "don't", "shouldn't", "couldn't",
             "won't", "can't", 'nobody', 'naught', 'nil', 'nix', 'nonentity', 'nullity', 'nothingness',
             'zilch']



language = 'greek'
iso639 = 'el'

spacy_model_sm = iso639 + '_core_news_sm'
spacy_model_md = iso639 + '_core_news_md'
spacy_model_lg = iso639 + '_core_news_lg'

sentiment_lexicon_class = None
pos_token_sentiment = None

negations = ['όχι', 'κανένας', 'ούτε', 'ποτέ', 'δεν', 'κανένας', 'νιξ', 'μη ουσία', 'ακυρότητα', 'τίποτα']


from scipy.cluster import hierarchy
import matplotlib.pyplot as plt
from compling.unsupervisedLearning.unsupervised_learning import UnsupervisedLearning
from typing import Callable, Union, Tuple


class Linkage(UnsupervisedLearning):
    """
    Hierarchical clustering is a method of cluster analysis which seeks to build a hierarchy of clusters.
    """

    def run(self, method: str = 'complete', metric: Union[str, Callable[[object], float]] = 'euclidean',
            optimal_ordering: bool = False) -> object:
        """
        Runs the hierarchical clustering for the input vectors

        Args:
            method (str, optional, default=complete): Methods for calculating the distance between the newly formed cluster u and each v. <br/> See https://docs.scipy.org/doc/scipy/reference/generated/scipy.cluster.hierarchy.linkage.html . <br/> Method can be ['complete', 'single', 'average', 'weighted', 'centroid', 'median', 'ward']
            metric (Union[str, Callable[[object], float]], optional, default=euclidean): Metricstr or function. <br/> The distance metric to use on the collection of vectors. <br/> See: https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.distance.pdist.html#scipy.spatial.distance.pdist for a list of valid distance metrics. A custom distance function can also be used.
            optimal_ordering (bool, optional, default=False): If True, the linkage matrix will be reordered so that the distance between successive leaves is minimal. <br/> This results in a more intuitive tree structure when the data are visualized. <br/> Default to False, because this algorithm can be slow, particularly on large datasets

        Returns:
            object: The hierarchical clustering encoded as a linkage matrix. <br/> See: https://docs.scipy.org/doc/scipy/reference/generated/scipy.cluster.hierarchy.linkage.html
        """

        self.dendogram = hierarchy.linkage(self.data.T, method)
        self.method = method

        return self.dendogram

    def plot(self, figsize: Tuple[int, int] = (10, 12), save: str = None,
             **kwargs) -> None:
        """
        Plots the hierarchical clustering result.

        Args:
            figsize (Tuple[int, int], optional, default=10,12): Figure size: width, height in inches.
            save (str, optional, default=None): Figure is not saved if save is None. <br/> Set it with a output name if you want save it.
            kwargs (object, optional): See: https://docs.scipy.org/doc/scipy/reference/generated/scipy.cluster.hierarchy.dendrogram.html .
        """

        if not 'labels' in kwargs:
            kwargs['labels'] = self.data.T.index
        if not 'leaf_rotation' in kwargs:
            kwargs['leaf_rotation'] = 0
        if not 'orientation' in kwargs:
            kwargs['orientation'] = 'right'
        if not 'leaf_font_size' in kwargs:
            kwargs['leaf_font_size'] = 12

        fig, ax = plt.subplots(figsize=figsize)

        dn = hierarchy.dendrogram(self.dendogram, ax=ax, **kwargs)

        plt.tight_layout()

        if save:
            fig1 = plt.gcf()
            plt.show()
            plt.draw()
            fig1.savefig(save)
        else:
            plt.show()

        plt.close()



language = 'japanese'
iso639 = 'ja'

spacy_model_sm = iso639 + '_core_news_sm'
spacy_model_md = iso639 + '_core_news_md'
spacy_model_lg = iso639 + '_core_news_lg'

sentiment_lexicon_class = None
pos_token_sentiment = None

negations = ['ない', 'ない', 'いいえ', 'なし', 'どちらでもない', 'どちらでもない', 'まったくない', 'しない', 'ない', 'しない',
       'すべきではない', 'できなかった', 'しない', 'できない', '誰もいない', 'ない', 'ジルチ」']



from compling.nlp import NLP
from compling.config import ConfigManager
from spacy.tokens.token import Token as SpacyToken
from typing import Dict
import compling.lexicalAnalysis.tokenization.token_abc as abctoken
from compling.descriptor import f2att


class Token(abctoken.Token):
    """
    Represents a generic **token**, which is a sequence of characters in the input.
    A token is identified using a tokenizer, which breaks up the sequence of characters in the input into a sequence
    of tokens.
    """

    def __init__(self,
                 token_id: int,
                 token: SpacyToken,
                 nlp: NLP,
                 config: ConfigManager,
                 sent_id: int,
                 para_id: int,
                 doc_id: int,
                 sent_pos: int,
                 para_pos: int,
                 doc_pos: int,
                 is_negated: bool,
                 metadata: dict) -> None:
        """
        Represents a generic **token**, which is a sequence of characters in the input.
        A token is identified using a tokenizer, which breaks up the sequence of characters in the input into a sequence
        of tokens.



        Args:
           token_id (int): Unique token identifier.
           token (SpacyToken): A SpacyToken object.
           nlp (NLP): A Natural Language Processing object.
           config (ConfigManager): A Config manager object.
           sent_id (int): Unique sentence identifier. The id of the sentence the token occurs in.
           para_id (int): Unique paragraph identifier. The id of the paragraph the token occurs in.
           doc_id (int): Unique document identifier. The id of the document the token occurs in.
           sent_pos (int): Position index of the token in the sentence.
           para_pos (int): Position index of the token in the paragraph.
           doc_pos (int): Position index of the token in the document.
           metadata (dict): Token metadata as a key-value pairs.

        Returns:
           None.
        """

        # unique identifiers
        self.token_id_ = token_id
        self.sent_id_ = sent_id
        self.para_id_ = para_id
        self.doc_id_ = doc_id

        # Spacy Token
        self.token = token

        # Natural Language Object
        self.nlp = nlp

        # Text
        self.text_ = self.text()

        # position index
        self.sent_pos_ = sent_pos
        self.para_pos_ = para_pos
        self.doc_pos_ = doc_pos

        # inversion polarity
        self.is_negated_ = is_negated

        # add some field only if the value in config.ini file is True
        for k, v in config.get_section(s='Token_record').items():
            if int(v):
                setattr(self, k + '_', getattr(self, k)())

        # add metadata for this token record
        self.add_metadata(metadata)

    @f2att
    def text(self) -> str:
        """Returns the text of the token."""

        # Avoid exceptions in mongodb
        text = self.token.text.replace('.', '\u002E')

        if not self.nlp.is_ngram(text) or text.endswith("__"):
            return text
        else:
            # elif self.nlp.is_ngram(text) and not text.endswith("__"):
            # e.g. ngram: __middle_east__
            # a token may be: __middle_east__ern
            # so --> __middle_eastern__
            return "__" + text.replace("__", "") + "__"

    @f2att
    def sentiment(self) -> Dict[str, float]:
        """
        Returns the token polarities. <br/>
        Each polarity value is between 0 and 1. <br/>
        _p_ in _[0,1]_

        Returns:
            Dict[str, float]: Polarities type as keys, sentiment values as values.

                    Example:
                    token.text="fuck"

                    >> token.sentiment()
                    {
                     'pos' : 0.0,
                     'neg' : 1.0,
                     'obj' : 0.0
                    }
        """

        if self.pos() not in ('ADJ', 'ADV', 'AUX', 'NOUN', 'PROPN', 'VERB'):
            return {'pos': 0.0, 'neg': 0.0, 'obj': 0.0}

        prefix = "NOT_"
        text = self.lemma()

        if self.is_negated():
            text = prefix + text

        return self.nlp.token_sentiment(text)

    def shape(self) -> str:
        """
        Returns a transformed token version to show orthographic features of the token.

        Alphabetic characters are replaced by x or X, and numeric characters are replaced by d, and sequences of the
        same character are truncated after length 4. For example, "Xxxx" or "dd".
        """
        return self.token.shape_

    def lower(self) -> str:
        """Returns the lowercase version of the token text."""
        return self.nlp.lower(self.text())

    @f2att
    def stem(self) -> str:
        """Returns the stem of the token text."""

        return self.nlp.stem(self.text())

    def lemma(self) -> str:
        """Returns the lemma of the token text."""
        return self.token.lemma_.lower()

    def pos(self) -> str:
        """Returns the pos tag of the token text."""
        return self.token.pos_

    def dep(self) -> str:
        """Returns the dep tag of the token text."""
        return self.token.dep_

    def lang(self) -> str:
        """Returns the language of the token."""
        return self.nlp.nlp_language.iso639

    @f2att
    def is_stopword(self) -> bool:
        """Returns True if the token text is a stopword else False."""

        return self.nlp.is_stopword(self.text())

    @f2att
    def is_ngram(self) -> bool:
        """Returns True if the token text is a ngram else False."""

        return self.nlp.is_ngram(self.text())

    def is_digit(self) -> bool:
        """Returns True if the token text is a digit else False."""
        return self.nlp.is_digit(self.text())

    def is_upper(self) -> bool:
        """Returns True if the token text is upper else False."""
        return self.text().isupper()

    def is_capitalize(self) -> bool:
        """Returns True if the token text is capitalize else False."""
        return self.nlp.is_capitalize(self.text())

    def is_punct(self) -> bool:
        """Returns True if the token text is a punctuation symbol else False."""
        return self.nlp.is_punct(self.text())

    def is_space(self) -> bool:
        """Returns True if the token text is a space character else False."""
        return self.nlp.is_space(self.text())

    def is_bracket(self) -> bool:
        """Returns True if the token text is a bracket symbol else False."""
        return self.token.is_bracket

    def is_ascii(self) -> bool:
        """Returns True if the token text is encoded in ascii else False."""
        return self.token.is_ascii

    def is_negated(self) -> bool:
        """Returns True if the token text is negated in the sentence."""
        return self.is_negated_

    def add_metadata(self, metadata: dict) -> None:
        """Adds metadata as attributes of the token."""

        for k, v in metadata.items():
            if hasattr(self, k + '_'):
                continue
            setattr(self, k + '_', v)

    def to_dict(self) -> dict:
        """Converts the token object to a dict record."""

        return {k: v for k, v in self.__dict__.items() if k.endswith('_')}



import gensim
from compling.embeddings.documents.dembeddings import DocumentEmbeddings
from typing import *


class Doc2vec(DocumentEmbeddings):
    """
    Doc2vec is an unsupervised algorithm to generate vectors for sentence/paragraphs/documents. The algorithm is an adaptation of word2vec which can generate vectors for words.

    The vectors generated by doc2vec can be used for tasks like finding similarity between sentences/paragraphs/documents.
    """
    def __init__(self, index: Iterable[dict], id_field:Union[str, List[str]], output:str=None, text_field='text', skipgram_ws: int = None):
        super().__init__(output=output, index=index, text_field=text_field, skipgram_ws=skipgram_ws, tag=True,
                         module=gensim.models.doc2vec.Doc2Vec, id_field=id_field)





import re
from tqdm import tqdm
import numpy as np
import matplotlib.pyplot as plt
from collections import defaultdict
from compling.analysis.sentiment.sentiment import SentimentLexiconBased
from typing import *
from wordcloud import WordCloud


class SentimentAnalyzer(SentimentLexiconBased):
    """
    The most popular unsupervised strategies used in sentiment analysis are lexical-based methods. They make use of a predefined list of words, where each word is associated with a specific sentiment. Lexicon-based strategies are very efficient and simple methods. They make use of a sentiment lexicon to assign a polarity value to each text document by following a basic algrithm. A sentiment lexicon is a list of lexical features (e.g., words, phrase, etc.) which are labeled according to their semantic orientation (i.e. polarity) as either positive or negative.

    A SentimentAnalyzer object allows to perform sentiment analysis through a lexicon-based approach.

    SentimentAnalyzer uses a summation _strategy_: the polarity level of a document is calculated as the sum of the polarities of all the words making up the
    document.

    The analysis detects negation pattern and reverses the negated tokens polarity.

    Providing a regex, you can filter sentences/paragraphs/documents to analyze.

    Providing a pos list and/or a dep list you can filter the words whose polarities will be summed.
    """

    def filter(self, index: Iterable[dict],
               regex_pattern: str,
               text_field: str = 'text') -> Iterable[dict]:
        """
        Filters the tokens/sentences/paragraphs/documents to be considered to calculate a sentiment value.

        Args:
            index (Iterable[dict]): Records of an index produced during the tokenization stage (e.g. sentences, paragraphs, documents).
            regex_pattern (str): The filter uses this regex.
            text_field (str, optional, default=text): For each record the regex will be evaluated on this field.
        Returns:
            Iterable[dict]: The record filtered.
        """


        text_field+='_'
        pattern = re.compile(regex_pattern)

        self.__filtered_id__ = defaultdict(list)

        # Index filtered grouped by group_by_field
        for record in tqdm(index, desc='Filtering in progress...', leave=True, position=0):
            if pattern.match(record[text_field]):

                # add id to __filtered_id_
                for k in record:
                    if '_id_' in k:
                        self.__filtered_id__[k].append(record[k])
                        break

                yield record

    def run(self, index: Iterable[dict],
            text_field: str='text',
            min_len: int = 0,
            lang: str = None,
            pos: Tuple[str, ...] = None,
            dep: Tuple[str, ...] = None) -> Dict[str, Dict[str, float]]:
        """
        Performs sentiment analysis of the texts in the index calculating the polarity level for each different
        group_by_field value.

        Args:
            index (Iterable[dict]): Records of an index produced during the tokenization stage (e.g. sentences, paragraphs, documents).

                    Example:
                    >> s = SentimentAnalyzer(group_by_field='author')
                    >> sentences_filtered = s.filtered(sentences, regex_pattern='.*family.*')
                    >> s.run(index=sentences_filtered)

            text_field (str, optional, default=text): For each record 'min_len' will be checked on this field.
            min_len (int, optional, default=0): The minimum length a text must have to be considered
            lang (str, optional, default=None): You can filter the sentiment analysis for the document of a specific language.
            pos (Tuple[str, ...], optional, default=None): If index records are tokens, part of speech of the tokens. <br/> Filters the most informative tokens: tokens whose PoS tag is in the 'pos' list. <br/> If None, the filter will be ignored.
            dep (Tuple[str, ...], optional, default=None): If index records are tokens, dependencies of the tokens. <br/> Filters the most informative tokens: tokens whose dep tag is in the 'dep' list. <br/> If None, the filter will be ignored.

        Returns:
            Dict[str, Dict[str, float]]: A dict contains the polarity value for each different grouped_by_field value.

                         Example:
                         {
                          "grouped_by_field1": {
                                                "pos": 0,
                                                "neg":1,
                                                "obj":3},
                           ...,
                         }
        """

        # Info about sentiment
        polarities = defaultdict(lambda: defaultdict(float))
        count = defaultdict(lambda: defaultdict(int))

        text_field = text_field+'_'


        for record in tqdm(index, desc='SentimentAnalysis in progress...', leave=True, position=0):

            n_token = len(record[text_field].split())
            if n_token == 1 and len(record[text_field]) < min_len or min_len > n_token and n_token > 1:
                continue

            if lang is not None and lang != record['lang_'] or \
                    pos is not None and record['pos_'] in pos or \
                    dep is not None and record['dep_'] in dep:
                continue

            key = " + ".join([record[g] for g in self.group_by_field]).replace('.', '\u002E')

            for k, v in record['sentiment_'].items():
                polarities[key][k] += v
                if v > 0:
                    count[key][k] += 1

        for k in polarities.keys():
            if count[k]['pos'] > 0 or count[k]['neg'] > 0:
                polarities[k]['pos'] /= max(count[k]['pos'], count[k]['neg'])
                polarities[k]['neg'] /= max(count[k]['pos'], count[k]['neg'])
            if count[k]['obj'] > 0 or count[k]['obj'] > 0:
                polarities[k]['obj'] /= count[k]['obj']


        return polarities

    def sentitokens(self,
                    tokens: Iterable[dict],
                    text_field:str='text',
                    min_len:int=0,
                    lang:str=None,
                    pos: Tuple[str, ...] = None,
                    dep: Tuple[str, ...] = None) -> dict:
        """

        Args:
            tokens (Iterable[dict]): Stream of token records produced during the Tokenization stage.
            text_field (str, optional, default=text): The text field of each token.
            min_len (int, optional, default=0): The minimum length a token must have to be considered
            lang (str, optional, default=None): You can filter the token of a specific language.
            pos (Tuple[str, ...], optional, default=None): Part of speech of the tokens. <br/> Filters the most informative tokens: tokens whose PoS tag is in the 'pos' list. <br/> If None, the filter will be ignored.
            dep (Tuple[str, ...], optional, default=None): Dependencies of the tokens. <br/> Filters the most informative tokens: tokens whose dep tag is in the 'dep' list. <br/> If None, the filter will be ignored.

        Returns:
            dict: A dict containing the words classified as positive/negative, as keys, with their frequency value, as values, for each
            different grouped_by_field value.

                         Example:
                         {
                          "grouped_by_field1": {
                                                "neg": {
                                                        "fuck": 12,
                                                        "bad": 7,
                                                        ...
                                                        },
                                                "pos": {
                                                        "good": 11,
                                                        "love": 2,
                                                        ...
                                                        }
                                                },
                          ...
                        }
        """

        # Info about sentiment
        polarities = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))

        text_field = text_field + '_'

        if hasattr(self, '__filtered_id__'):
            key_id_filtered = list(self.__filtered_id__.keys())[0]
        else:
            key_id_filtered = None

        for record in tqdm(tokens, desc='Sentitokens classification in progress...', leave=True, position=0):

            if len(record[text_field]) < min_len:
                continue

            if lang is not None and lang!=record['lang_'] or\
                pos is not None and record['pos_'] in pos or \
                dep is not None and record['dep_'] in dep or \
                    key_id_filtered is not None and record[key_id_filtered] not in self.__filtered_id__[key_id_filtered]:
                continue

            key = " + ".join([record[g] for g in self.group_by_field]).replace('.', '\u002E')

            if 'is_negated_' in record and record['is_negated_']:
                token = 'NOT_'+record[text_field]
            else:
                token = record[text_field]

            if record['sentiment_']['pos'] > record['sentiment_']['neg']:
                polarities[key]['pos'][token] += 1
            elif record['sentiment_']['neg'] > record['sentiment_']['pos']:
                polarities[key]['neg'][token] += 1

        return polarities

    def plot(self, polarities, ylabel: str = None, xlabel: str = None, figsize: Tuple[int, int] = (12, 8),
             title: str = None, hidexticks: bool = True, save: bool = None) -> None:
        """
        Draws a scatter plot of positive and negative polarities level for each group_by_field sort by time (If date_field is not None).

        Args:
            xlabel (str, optional, default=None): Label on x-axis.
            ylabel (str, optional, default=None): Label on y-axis.
            figsize (Tuple[int, int], optional, default=12,8): Figure size: width, height in inches.
            title (str, optional, default=None): If not None, draws title string on the figure.
            hidexticks (bool, optional, default=False): If True, hides xticks.
            save (str, optional, default=None): Figure is not saved if save is None. <br/> Set it with an output name if you want save it.
        """

        plt.figure(figsize=figsize)
        plt.plot()

        polarities_ = {"pos": {"color": "red", "label": "Positive (+)"},
                      "neg": {"color": "black", "label": "Negative (-)"}}

        for polarity in polarities_:
            plt.tight_layout()
            x, y = list(polarities.keys()), [polarities[id_vector][polarity] for id_vector in polarities]
            plt.plot(x, y, label=polarities_[polarity]["label"], color=polarities_[polarity]["color"])
            plt.xticks(x, rotation="vertical")
            plt.title('\n{}'.format(title), size=20)
            if not ylabel is None:
                plt.ylabel('{}\n'.format(ylabel), size=20)
            if not xlabel is None:
                plt.xlabel('\n{}'.format(xlabel), size=20)

        plt.legend()  # bbox_to_anchor=(1.5, 0.5), loc='right', ncol=1)

        if not hidexticks:
            plt.xticks([])

        if save:
            plt.tight_layout()
            fig1 = plt.gcf()
            plt.show()
            plt.draw()
            fig1.savefig(save)
        else:
            plt.show()

    def circleplot(self, polarities:Dict[str, Dict[str, float]], radius: int = 3, figsize: Tuple[int, int] = (10, 10), title: str = None,
                   width_modifier: float = 1.3,
                   offset: float = 0.8, save: bool = False) -> None:
        """
        Draws a circle plot of positive and negative polarities level for each group_by_field sort by time (If date_field is not None).

        Args:
            polarities (Dict[str, Dict[str, float]]): The result of the method run.
            radius (int, optional, default=3): The radius of the circle.
            figsize (Tuple[int, int], optional, default=10,10): Figure size: width, height in inches.
            width_modifier (float, optional, default=1.3): Width of the bars.
            offset (float, optional, default=0.8): Offset for the label on the bars.
            save (str, optional, default=None): Figure is not saved if save is None. <br/> Set it with an output name if you want save it.
            title (str, optional, default=None):  If not None, draws title string on the figure.
        """

        data = polarities

        plt.figure(figsize=figsize)
        ax = plt.subplot(111, polar=True)
        ax.axis('off')

        N = len(data)
        radius = radius

        theta = np.linspace(0.0, 2 * np.pi, N, endpoint=False)
        width = (width_modifier * np.pi) / N

        pos_values, neg_values = np.array([data[i]["pos"] for i in data]), np.array([data[i]["neg"] for i in data])

        pos = ax.bar(theta, pos_values, width=width, bottom=radius)
        neg = ax.bar(theta, -neg_values, width=width, bottom=radius)

        for p, n in zip(pos, neg):
            p.set_facecolor("red")
            p.set_alpha(0.8)
            n.set_facecolor("black")
            n.set_alpha(0.8)

        # labeling
        for i, bar in enumerate(pos):
            t = theta[i] + 0.002
            x, y = (bar.get_x() + width / 2, offset * len(list(data)[i]) + bar.get_height())
            if np.cos(t) < 0: t = t - np.pi
            ax.text(x, y, list(data)[i], rotation=np.rad2deg(t), ha="center", va="center")

        if title:
            plt.title(title)

        if save:
            plt.tight_layout()
            fig1 = plt.gcf()
            plt.show()
            plt.draw()
            fig1.savefig(save)
        else:
            plt.show()

    def wordcloud(self, words:Dict[str, Dict[str, int]],
                  color: str = 'black', title: str = None, figsize: Tuple[int, int] = (10, 8),
                  save=None, width: int = 1000, height: int = 1000,
                  max_words: int = 100) -> None:
        """
        Draws a word Cloud with the words of grouped_by_field_value.

        Args:
            words (Dict[str, Dict[str, float]]): The sentitokens for a specific key and a specific polarity ['pos'/'neg'].
            color (str, optional, default=black): Background color.
            title (str, optional, default=None): If not None, draws title string on the figure.
            figsize (Tuple[int, int], optional, default=10,10): Figure size: width, height in inches.
            save (str, optional, default=None): Figure is not saved if save is None. Set it with a output name if you want save it.
            width (int, optional, default=1000): Width of the canvas.
            height (int, optional, default=1000): Height of the canvas.
            max_words (int, optional, default=100): The maximum number of words.
        """



        sentiment = words
        wordcloud = WordCloud(width=width, background_color=color, height=height,
                              max_words=max_words).generate_from_frequencies(sentiment)
        plt.figure(figsize=figsize)
        plt.imshow(wordcloud, label=title)
        plt.title(title, fontsize=22)

        if save:
            plt.tight_layout()
            fig1 = plt.gcf()
            plt.show()
            plt.draw()
            fig1.savefig(save)
        else:
            plt.show()

        plt.close()



import abc
from typing import *
from compling.nlp import NLP


class Vectorization(metaclass=abc.ABCMeta):
    """
    The process of converting text into vector is called vectorization.
    The set of corpus documents vectorized corpus makes up the Vector Space Model, which can have a sparse or
    dense representation.
    """

    def __init__(self, group_by_field: Union[str, List[str]], token_field: str = 'text') -> None:
        """
        **\_\_init\_\_**: Creates a new Vectorization object.

        Args:
            token_field (str, optional, default=text): A field containing the text of tokens (e.g. text, lemma, ...). <br/> The tokens will be grouped by this field.
            group_by_field (Union[str, List[str]]): Fields linked to tokens in the corpus (e.g. doc_id, sent_id, metadata1, ...).  <br/> The tokens will be grouped by these field.
        """
        self.token_field = token_field + '_'

        if isinstance(group_by_field, str):
            self.group_by_field = [group_by_field + '_']
        else:
            self.group_by_field = [g + '_' for g in group_by_field]

        self.group_by_field_key = " + ".join(self.group_by_field)

        self.nlp = NLP()

    @abc.abstractmethod
    def tf(self, tokens: Iterable[dict], pos: Tuple[str, ...] = ("NOUN", "VERB", "PROPN", "NGRAM"),
           dep: Tuple[str, ...] = None, lang: str = None, min_len: int = 0,
           min_count: int = 0, format_: str = "docterm-matrix", boolean=False) -> Union[
        Dict[str, dict], Iterable[dict]]:
        """
        Calculates Term Frequency.

        Args:
            tokens (Iterable[dict]): Stream of token records produced during the Tokenization stage.

                    Example:
                    >>> token = next(tokens)
                    >>> print(token)
                    {
                    "token_id_": 0,
                    "sent_id_": 0,
                    "para_id_": 0,
                    "doc_id_": 0,
                    "text_": "When",
                    "sent_pos_": 0,
                    "para_pos_": 0,
                    "doc_pos_": 0,
                    "lower_": "when",
                    ...
                    }

            pos (Tuple[str, ...], optional, default=NOUN,VERB,PROPN,ADJ): Part of speech of the tokens. <br/> Filters the most informative tokens: tokens whose PoS tag is in the 'pos' list. <br/> If None, the filter will be ignored.
            dep (Tuple[str, ...], optional, default=None): Dependencies of the tokens. <br/> Filters the most informative tokens: tokens whose dep tag is in the 'dep' list. <br/> If None, the filter will be ignored.
            lang (str, optional, default=None): Filters the tokens of the language (iso639). <br/> If None, the filter will be ignored.
            min_len (int, optional, default=0): The minimum length a token must have to be considered.
            min_count (int, optional, default=0): The minimum frequency a token must have in the corpus to be considered.
            format_ (str, optional, default=docterm-matrix): Format of the vectorization records.

                Valid input values:
                ["postings-list", "docterm-matrix"]

            boolean (bool, optional, default=False): If True return Boolean Vectors else Term Frequency. <br/> If True, the result is the same as onehot method.

        Returns:
            Union[Dict[str, dict], Iterable[dict]]: If format_ == 'postings-list', returns a postings list; else a docterm matrix.

                    Example:
                    [format_ == "postings-list"]

                    {
                        "token1":
                              {
                                "doc1": tf1,
                                ...
                              },
                        ...
                    }

                     Example:
                    [format_ == "docterm-matrix"]

                    [{
                        "doc": "doc1",
                        "bow":
                              {
                               token1: tf1,
                               ...
                               }
                    },
                    ...
                    ]
        """

    @abc.abstractmethod
    def tfidf(self, tokens: Iterable[dict], normalize=True,
              pos: Tuple[str, ...] = ("NOUN", "VERB", "PROPN", "NGRAM"),
              dep: Tuple[str, ...] = None, lang: str = None, min_len: int = 0,
              min_count: int = 0, format_: str = 'docterm-matrix') -> Union[Dict[str, dict], Iterable[dict]]:
        """
        Calculates Term Frequency Inverse Document Frequency.


        Args:
            tokens (Iterable[dict]): Stream of token records produced during the Tokenization stage.

                    Example:
                    >>> token = next(tokens)
                    >>> print(token)
                    {
                    "token_id_": 0,
                    "sent_id_": 0,
                    "para_id_": 0,
                    "doc_id_": 0,
                    "text_": "When",
                    "sent_pos_": 0,
                    "para_pos_": 0,
                    "doc_pos_": 0,
                    "lower_": "when",
                    ...
                    }

            normalize (bool, optional, default=True): If True, normalizes TF for max(TF).
            pos (Tuple[str, ...], optional, default=NOUN,VERB,PROPN,ADJ): Part of speech of the tokens. <br/> Filters the most informative tokens: tokens whose PoS tag is in the 'pos' list. <br/> If None, the filter will be ignored.
            dep (Tuple[str, ...], optional, default=None): Dependencies of the tokens. <br/> Filters the most informative tokens: tokens whose dep tag is in the 'dep' list. <br/> If None, the filter will be ignored.
            lang (str, optional, default=None): Filters the tokens of the language (iso639). <br/> If None, the filter will be ignored.
            min_len (int, optional, default=0): The minimum length a token must have to be considered.
            min_count (int, optional, default=0): The minimum frequency a token must have in the corpus to be considered.
            format_ (str, optional, default=docterm-matrix): Format of the vectorization records.

                    Valid input values:
                    ["postings-list", "docterm-matrix"]

        Returns:
            Union[Dict[str, dict], Iterable[dict]]: If format_ == 'postings-list', returns a postings list; else a docterm matrix.

                    Example:
                    [format_ == "postings-list"]

                    {
                        "token1":
                              {
                                "group_by_field1": tfid1,
                                ...
                              },
                        ...
                    }

                     Example:
                    [format_ == "docterm-matrix"]

                    [{
                        group_by_field: "group_by_field1",
                        "bow":
                              {
                               token1: tfid1,
                               ...
                               }
                    },
                    ...
                    ]
        """

    @abc.abstractmethod
    def mi(self, tokens: Iterable[dict], pos: Tuple[str, ...] = ("NOUN", "VERB", "PROPN", "NGRAM"),
           dep: Tuple[str, ...] = None, lang: str = None, min_len: int = 0,
           min_count: int = 0, format_: str = 'docterm-matrix') -> Union[Dict[str, dict], Iterator[dict]]:
        """
        Calculates Mutual Information.

        Args:
            tokens (Iterable[dict]): Stream of token records produced during the Tokenization stage.

                    Example:
                    >>> token = next(tokens)
                    >>> print(token)
                    {
                    "token_id_": 0,
                    "sent_id_": 0,
                    "para_id_": 0,
                    "doc_id_": 0,
                    "text_": "When",
                    "sent_pos_": 0,
                    "para_pos_": 0,
                    "doc_pos_": 0,
                    "lower_": "when",
                    ...
                    }

            pos (Tuple[str, ...], optional, default=NOUN,VERB,PROPN,ADJ): Part of speech of the tokens. <br/> Filters the most informative tokens: tokens whose PoS tag is in the 'pos' list. <br/> If None, the filter will be ignored.
            dep (Tuple[str, ...], optional, default=None): Dependencies of the tokens. <br/> Filters the most informative tokens: tokens whose dep tag is in the 'dep' list. <br/> If None, the filter will be ignored.
            lang (str, optional, default=None): Filters the tokens of the language (iso639). <br/> If None, the filter will be ignored.
            min_len (int, optional, default=0): The minimum length a token must have to be considered.
            min_count (int, optional, default=0): The minimum frequency a token must have in the corpus to be considered.
            format_ (str, optional, default=docterm-matrix): Format of the vectorization records.

                    Valid input values:
                    ["postings-list", "docterm-matrix"]

        Returns:
            Union[Dict[str, dict], Iterable[dict]]: If format_ == 'postings-list', returns a postings list; else a docterm matrix.

                    Example:
                    [format_ == "postings-list"]

                    {
                        "token1":
                              {
                                "doc1": mi1,
                                ...
                              },
                        ...
                    }

                    Example:
                    [format_ == "docterm-matrix"]

                    [{
                        "doc": "doc1",
                        "bow":
                              {
                               token1: mi1,
                               ...
                               }
                    },
                    ...
                    ]
        """

    @abc.abstractmethod
    def onehot(self, tokens: Iterable[dict], pos: Tuple[str, ...] = ("NOUN", "VERB", "PROPN", "NGRAM"),
               dep: Tuple[str, ...] = None, lang: str = None, min_len: int = 0,
               min_count: int = 0, format_: str = 'docterm-matrix') -> Union[Dict[str, dict], Iterable[dict]]:
        """
        Calculates One-hot encoding.

        Args:
            tokens (Iterable[dict]): Stream of token records produced during the Tokenization stage.

                    Example:
                    >>> token = next(tokens)
                    >>> print(token)
                    {
                    "token_id_": 0,
                    "sent_id_": 0,
                    "para_id_": 0,
                    "doc_id_": 0,
                    "text_": "When",
                    "sent_pos_": 0,
                    "para_pos_": 0,
                    "doc_pos_": 0,
                    "lower_": "when",
                    ...
                    }

            pos (Tuple[str, ...], optional, default=NOUN,VERB,PROPN,ADJ): Part of speech of the tokens. <br/> Filters the most informative tokens: tokens whose PoS tag is in the 'pos' list. <br/> If None, the filter will be ignored.
            dep (Tuple[str, ...], optional, default=None): Dependencies of the tokens. <br/> Filters the most informative tokens: tokens whose dep tag is in the 'dep' list. <br/> If None, the filter will be ignored.
            lang (str, optional, default=None): Filters the tokens of the language (iso639). <br/> If None, the filter will be ignored.
            min_len (int, optional, default=0): The minimum length a token must have to be considered.
            min_count (int, optional, default=0): The minimum frequency a token must have in the corpus to be considered.
            format_ (str, optional, default=docterm-matrix): Format of the vectorization records.

                    Valid input values:
                    ["postings-list", "docterm-matrix"]

        Returns:
            Union[Dict[str, dict], Iterable[dict]]: If format_ == 'postings-list', returns a postings list; else a docterm matrix.

                    Example:
                    [format_ == "postings-list"]

                    {
                        "token1":
                              {
                                "doc1": 1,
                                "doc2": 0,
                                ...
                              },
                        ...
                    }

                    Example:
                    [format_ == "docterm-matrix"]

                    [{
                        "doc": "doc1",
                        "bow":
                              {
                               token1: 1,
                               ...
                               }
                    },
                    ...
                    ]
        """

    @abc.abstractmethod
    def run(self, mode: str, tokens: Iterable[dict], normalize=True,
            pos: Tuple[str, ...] = ("NOUN", "VERB", "PROPN", "NGRAM"),
            dep: Tuple[str, ...] = None,
            lang: str = None, min_len=0, min_count=0,
            metadata: Dict[str, dict] = None) -> Iterable[dict]:
        """
        Runs the vectorization grouping token by self.token_field and self.group_by_field.

        Args:
            mode (str): Vectorization mode. <br/>

                    Valid input values:
                    ["tf", "tfidf", "mi", "onehot"]

            tokens (Iterable[dict]): Stream of token records produced during the Tokenization stage.

                    Example:
                    >>> token = next(tokens)
                    >>> print(token)
                    {
                    "token_id_": 0,
                    "sent_id_": 0,
                    "para_id_": 0,
                    "doc_id_": 0,
                    "text_": "When",
                    "sent_pos_": 0,
                    "para_pos_": 0,
                    "doc_pos_": 0,
                    "lower_": "when",
                    ...
                    }

            normalize (bool, optional, default=True): If True, normalizes TF for max(TF). <br/> This parameter is ignored if mode in ['tf', 'onehot', 'mi'].
            pos (Tuple[str, ...], optional, default=NOUN,VERB,PROPN,ADJ): Part of speech of the tokens. <br/> Filters the most informative tokens: tokens whose PoS tag is in the 'pos' list. <br/> If None, the filter will be ignored.
            dep (Tuple[str, ...], optional, default=None): Dependencies of the tokens. <br/> Filters the most informative tokens: tokens whose dep tag is in the 'dep' list. <br/> If None, the filter will be ignored.
            lang (str, optional, default=None): Filters the tokens of the language (iso639). <br/> If None, the filter will be ignored.
            min_len (int, optional, default=0): The minimum length a token must have to be considered.
            min_count (int, optional, default=0): The minimum frequency a token must have in the corpus to be considered.
            metadata (Dict[str, dict], optional, default=None): A dict containing metadata for each different group_by_field value.  <br/> Each vector will be associated with its own metadata.

                    Example:

                    {
                     "group_by_field1": {
                                         metadata1: 'metadata1',
                                         metadata2: 'metadata2',
                                         ...
                                        },
                     "group_by_field2": {
                                          metadata1: 'metadata11',
                                          metadata2: 'metadata22',
                                         ...
                                        },
                    ...
                    }


        Returns:
            Iterable[dict]: Returns a docterm matrix.

                    Example:
                    >> v = Vectorizer('doc_id', 'text')
                    >> record = v.run('tfidf', tokens)
                    >> print(next(record))
                    {
                        "doc": "doc1",
                        "bow":
                              {
                               token1: frequency1,
                               ...
                               }
                        metadata1: "metadata1",
                        metadata2: "metadata2",
                        ...
                    }
        """



from tqdm import tqdm
import copy
import math
from typing import *
from collections import defaultdict
from compling.lexicalAnalysis.vectorization.nevectorization import NEVectorization

class NEVectorizer(NEVectorization):
    """
    The process of converting text into vector is called vectorization.
    The set of corpus documents vectorized corpus makes up the Vector Space Model, which can have a sparse or
    dense representation.

    A NEVectorization object allows you to create vectors grouping named_entities for arbitrary fields.

    Grouping named_entities by:

       * _doc_id_: you 're creating document vectors;
       * _sent_id_: you 're creating sentence vectors;
       * _author_: you're creating author vectors (each named_entities must have an 'author' field);
       ...

    You can group by multiple fields.

    It offers several functions to set the vector components values, such as:

       * **One-hot encoding** (binary values 0/1);
       * **Tf** (Term Frequency);
       * **TfIdf** (Term Frequency Inverse Document Frequency);
       * **MI** (Mutual Information)

    You can specify the vectorization representation format: Term x Document matrix, Postings list.
    """

    def postings2docterm(self, postings: Dict[str, dict]) -> Iterable[dict]:
        """
        Converts a postings list to docterm matrix.

        Args:
            postings (Dict[str, dict]): Postings list.

                    Example:
                    >> print(postings)
                    {
                     "ne1":
                           {
                             "doc1": frequency1,
                             ...
                           },
                     ...
                     }

        Returns:
            Iterable[dict]: Docterm matrix.

                    Example:
                    >> print(next(postings2docterm(postings)))
                    {
                     "doc": "doc1",
                     "bow":
                           {
                            ne1: frequency1,
                            ...
                            }
                     }
        """

        docterm = defaultdict(dict)
        for named_entity in postings:
            for doc, relevance in postings[named_entity].items():
                docterm[doc].update({named_entity: relevance})

        for doc in docterm:
            yield {self.group_by_field_key: doc, 'bow': docterm[doc]}

    def docterm2postings(self, docterm: Iterable[dict]) -> Dict[str, dict]:
        """
        Converts a docterm matrix to a postings list.

        Args:
            docterm (Iterator[dict]): Docterm matrix.

                    Example:
                    >> print(next(docterm))
                    {
                     "doc": "doc1",
                     "bow":
                           {
                            ne1: frequency1,
                            ...
                            }
                     }
                     # bow: Bag of Words
        Returns:
            Dict[str, dict]: Postings list.

                    Example:
                    >> print(docterm2postings(docterm))
                    {
                     "ne1":
                           {
                             "doc1": frequency1,
                             ...
                           },
                     ...
                     }
        """

        postings = defaultdict(dict)
        for record in docterm:
            for named_entity, relevance in record['bow'].items():
                postings[named_entity].update({record[self.group_by_field_key]: relevance})

        return postings

    def tf(self, index: Iterable[dict],
           min_len: int = 0,
           min_count: int = 0,
           format_: str = "docterm-matrix",
           boolean=False) -> Union[Dict[str, dict], Iterable[dict]]:
        """
        Calculates Term Frequency.

        Args:
            index (Iterable[dict]): Records of an index produced during the tokenization stage (e.g. sentences, paragraphs, documents).

                    Example:
                    >>> sentence = next(sentences)
                    >>> print(sentence)
                    {
                    "sent_id_": 0,
                    "text_": "When We , though all unworthy , were called to succeed on the Apostolic Throne the meek Pius X [...]",
                    "lemma_": "when -PRON- , though all unworthy , be call to succeed on the Apostolic Throne the meek Pius X [ ... ]",
                    [...]
                    "named_entities_": [{"text_": "Apostolic Throne", "start":12, "sent_pos_end_":13}, {"text_": "Pius X", "start": 16, "sent_pos_end_":17}],
                    ...
                    }

            min_len (int, optional, default=0): The minimum length a named_entity must have to be considered. A named_entity must be at least min_len characters long.
            min_count (int, optional, default=0): The minimum frequency a named_entity must have in the corpus to be considered.
            format_ (str, optional, default=docterm-matrix): Format of the vectorization records.

                    Valid input values:
                    ["postings-list", "docterm-matrix"]

            boolean (bool, optional, default=False): If True return Boolean Vectors else Term Frequency. <br/> If True, the result is the same as onehot method.

        Returns:
            Union[Dict[str, dict], Iterable[dict]]: If format_ == 'postings-list', returns a postings list; else a docterm matrix.

                    Example:
                    [format_ == "postings-list"]

                    {
                        "named_entity1":
                              {
                                "doc1": tf1,
                                ...
                              },
                        ...
                    }

                     Example:
                    [format_ == "docterm-matrix"]

                    [{
                        "doc": "doc1",
                        "bow":
                              {
                               named_entity1: tf1,
                               ...
                               }
                    },
                    ...
                    ]
        """

        if not hasattr(self, 'tf_postings'):

            def __filter__(__ne__):
                if len(__ne__) < min_len:
                    return False
                return True

            postings = defaultdict(lambda: defaultdict(int))
            for doc in tqdm(index, desc="TF calculation in progress...", position=0, leave=True):

                key = " + ".join([doc[g] for g in self.group_by_field]).replace('.', '\u002E')

                for ne in doc[self.ne_field]:
                    if len(ne['text_']) < min_len:
                        continue
                    else:
                        named_entity = ne["text_"]

                        if boolean:
                            postings[named_entity][key] = 1
                        else:
                            postings[named_entity][key] += 1

            for named_entity in copy.deepcopy(postings):
                if sum(postings[named_entity].values()) < min_count:
                    del postings[named_entity]

            self.tf_postings = postings

        if format_ == 'postings-list':
            return self.tf_postings
        else:
            return self.postings2docterm(self.tf_postings)

    def reset_tf(self):
        """If present, removes the tf postings list from memory.
         It will be re-calculated on the next run call."""
        if hasattr(self, 'tf_postings'):
            delattr(self, 'tf_postings')

    def tfidf(self,
              index: Iterable[dict],
              normalize=True,
              min_len=0,
              min_count: int = 0,
              format_: str = 'docterm-matrix') -> Union[Dict[str, dict], Iterable[dict]]:
        """
        Calculates Term Frequency Inverse Document Frequency.


        Args:
            index (Iterable[dict]): Records of an index produced during the tokenization stage (e.g. sentences, paragraphs, documents).

                    Example:
                    >>> sentence = next(sentences)
                    >>> print(sentence)
                    {
                    "sent_id_": 0,
                    "text_": "When We , though all unworthy , were called to succeed on the Apostolic Throne the meek Pius X [...]",
                    "lemma_": "when -PRON- , though all unworthy , be call to succeed on the Apostolic Throne the meek Pius X [ ... ]",
                    [...]
                    "named_entities_": [{"text_": "Apostolic Throne", "start":12, "sent_pos_end_":13}, {"text_": "Pius X", "start": 16, "sent_pos_end_":17}],
                    ...
                    }

            normalize (bool, optional, default=True): If True, normalizes TF for max(TF).
            min_len (int, optional, default=0): The minimum length a named_entity must have to be considered. A named_entity must be at least min_len characters long.
            min_count (int, optional, default=0): The minimum frequency a named_entity must have in the corpus to be considered.
            format_ (str, optional, default=docterm-matrix): Format of the vectorization records.

                    Valid input values:
                    ["postings-list", "docterm-matrix"]

        Returns:
            Union[Dict[str, dict], Iterable[dict]]: If format_ == 'postings-list', returns a postings list; else a docterm matrix.

                    Example:
                    [format_ == "postings-list"]

                    {
                        "named_entity1":
                              {
                                "doc1": tfidf1,
                                ...
                              },
                        ...
                    }

                     Example:
                    [format_ == "docterm-matrix"]

                    [{
                        "doc": "doc1",
                        "bow":
                              {
                               named_entity1: tfidf1,
                               ...
                               }
                    },
                    ...
                    ]
        """

        postings = self.tf(index=index, min_len=min_len, min_count=min_count, format_='postings-list')

        docterm = self.postings2docterm(postings)
        max_tf = {}
        n = 0
        for doc in docterm:
            n += 1
            max_tf[doc[self.group_by_field_key]] = max(doc['bow'].values())

        # Corpus size: the number of documents in the corpus
        N = n

        # The tf_dict will be update to tfidf_dict
        tfidf_postings = copy.deepcopy(postings)

        bar = tqdm(total=len(tfidf_postings), desc="TFIDF calculation in progress... ", position=0, leave=True)

        # For each named_entity in the postings list
        for named_entity in tfidf_postings.keys():

            docs = tfidf_postings[named_entity]

            # For each doc which the named_entity occurs in
            for doc, tf in docs.items():
                if normalize:
                    tf = tf / max_tf[doc]

                # Document Frequency
                df = len(postings[named_entity])

                # Inverse Document Frequency
                idf = math.log(N / df)

                # Term Frequency Inverse Document Frequency
                tfidf_postings[named_entity].update({doc: tf * idf})
            bar.update(1)
        bar.close()

        if format_ == 'postings-list':
            return tfidf_postings
        else:
            return self.postings2docterm(tfidf_postings)

    def mi(self,
           index: Iterable[dict],
           min_len: int = 0,
           min_count: int = 0,
           format_: str = 'docterm-matrix') -> Union[Dict[str, dict], Iterator[dict]]:
        """
        Calculates Mutual Information.

        Args:
            index (Iterable[dict]): Records of an index produced during the tokenization stage (e.g. sentences, paragraphs, documents).

                    Example:
                    >>> sentence = next(sentences)
                    >>> print(sentence)
                    {
                    "sent_id_": 0,
                    "text_": "When We , though all unworthy , were called to succeed on the Apostolic Throne the meek Pius X [...]",
                    "lemma_": "when -PRON- , though all unworthy , be call to succeed on the Apostolic Throne the meek Pius X [ ... ]",
                    [...]
                    "named_entities_": [{"text_": "Apostolic Throne", "start":12, "sent_pos_end_":13}, {"text_": "Pius X", "sent_pos_start_": 16, "sent_pos_end_":17}],
                    ...
                    }

            min_len (int, optional, default=0): The minimum length a named_entity must have to be considered. A named_entity must be at least min_len characters long.
            min_count (int, optional, default=0): The minimum frequency a named_entity must have in the corpus to be considered.
            format_ (str, optional, default=docterm-matrix): Format of the vectorization records.

                    Valid input values:
                    ["postings-list", "docterm-matrix"]

        Returns:
            Union[Dict[str, dict], Iterable[dict]]: If format_ == 'postings-list', returns a postings list; else a docterm matrix.

                    Example:
                    [format_ == "postings-list"]

                    {
                        "named_entity1":
                              {
                                "doc1": mi1,
                                ...
                              },
                        ...
                    }

                     Example:
                    [format_ == "docterm-matrix"]

                    [{
                        "doc": "doc1",
                        "bow":
                              {
                               named_entity1: mi1,
                               ...
                               }
                    },
                    ...
                    ]
        """

        postings = self.tf(index, min_len, min_count, format_='postings-list')
        docterm = {record[self.group_by_field_key]: record for record in self.postings2docterm(postings)}

        # x: group_by_field label
        # y: named_entity
        n_named_entities = sum([sum(postings[named_entity].values()) for named_entity in postings])

        mi_postings = defaultdict(lambda: defaultdict(int))
        for named_entity in tqdm(postings, desc="Mutual Information calculation in progress...", position=0, leave=True):
            posting = defaultdict(int, postings[named_entity])
            for label in posting:
                row = docterm[label]['bow']
                len_x = sum(posting.values())
                len_y = sum(row.values())

                p_x = len_x / n_named_entities
                p_y = len_y / n_named_entities
                p_x_y = posting[label] / n_named_entities

                mi_postings[named_entity][label] = p_x_y * math.log(p_x_y / (p_x * p_y))

        if format_ == 'postings-list':
            return mi_postings
        else:
            return self.postings2docterm(mi_postings)

    def onehot(self, index: Iterable[dict], min_len: int = 0,
               min_count: int = 0, format_: str = 'docterm-matrix') -> Union[Dict[str, dict], Iterable[dict]]:
        """
        Calculates One-hot encoding.

        Args:
            index (Iterable[dict]): Records of an index produced during the tokenization stage (e.g. sentences, paragraphs, documents).

                    Example:
                    >>> sentence = next(sentences)
                    >>> print(sentence)
                    {
                    "sent_id_": 0,
                    "text_": "When We , though all unworthy , were called to succeed on the Apostolic Throne the meek Pius X [...]",
                    "lemma_": "when -PRON- , though all unworthy , be call to succeed on the Apostolic Throne the meek Pius X [ ... ]",
                    [...]
                    "named_entities_": [{"text_": "Apostolic Throne", "sent_pos_start_":12, "sent_pos_end_":13}, {"text_": "Pius X", "sent_pos_start_": 16, "sent_pos_end_":17}],
                    ...
                    }

            min_len (int, optional, default=0): The minimum length a named_entity must have to be considered. A named_entity must be at least min_len characters long.
            min_count (int, optional, default=0): The minimum frequency a named_entity must have in the corpus to be considered.
            format_ (str, optional, default=docterm-matrix): Format of the vectorization records.

                    Valid input values:
                    ["postings-list", "docterm-matrix"]

        Returns:
            Union[Dict[str, dict], Iterable[dict]]: If format_ == 'postings-list', returns a postings list; else a docterm matrix.

                    Example:
                    [format_ == "postings-list"]

                    {
                        "named_entity1":
                              {
                                "doc1": 0/1,
                                ...
                              },
                        ...
                    }

                     Example:
                    [format_ == "docterm-matrix"]

                    [{
                        "doc": "doc1",
                        "bow":
                              {
                               named_entity1: 0/1,
                               ...
                               }
                    },
                    ...
                    ]
        """
        return self.tf(index, min_len, min_count, boolean=True, format_=format_)

    def run(self, mode: str, index: Iterable[dict], normalize=True,
            min_len=0, min_count=0,
            metadata: Dict[str, dict] = None) -> Iterable[dict]:
        """
        Runs the vectorization grouping named_entities by self.ne_field and self.group_by_field.

        Args:
            mode (str): Vectorization mode. <br/>

                    Valid input values:
                    ["tf", "tfidf", "mi", "onehot"]

            index (Iterable[dict]): Records of an index produced during the tokenization stage (e.g. sentences, paragraphs, documents).

                    Example:
                    >>> sentence = next(sentences)
                    >>> print(sentence)
                    {
                    "sent_id_": 0,
                    "text_": "When We , though all unworthy , were called to succeed on the Apostolic Throne the meek Pius X [...]",
                    "lemma_": "when -PRON- , though all unworthy , be call to succeed on the Apostolic Throne the meek Pius X [ ... ]",
                    [...]
                    "named_entities_": [{"text_": "Apostolic Throne", "sent_pos_start_":12, "sent_pos_end_":13}, {"text_": "Pius X", "sent_pos_start_": 16, "sent_pos_end_":17}],
                    ...
                    }

            normalize (bool, optional, default=True): If True, normalizes TF for max(TF). <br/> This parameter is ignored if mode in ['tf', 'onehot', 'mi'].
            min_len (int, optional, default=0): The minimum length a named_entity must have to be considered. A named_entity must be at least min_len characters long.
            min_count (int, optional, default=0): The minimum frequency a named_entity must have in the corpus to be considered.
            metadata (Dict[str, dict], optional, default=None): A dict containing metadata for each different group_by_field value.  <br/> Each vector will be associated with its own metadata.

                    Example:

                    {
                     "group_by_field1": {
                                         metadata1: 'metadata1',
                                         metadata2: 'metadata2',
                                         ...
                                        },
                     "group_by_field2": {
                                          metadata1: 'metadata11',
                                          metadata2: 'metadata22',
                                         ...
                                        },
                    ...
                    }


        Returns:
            Iterable[dict]: Returns a docterm matrix.

                    Example:
                    >> v = NEVectorizer('doc_id', 'text')
                    >> record = v.run('tfidf', documents)
                    >> print(next(record))
                    {
                        "doc": "doc1",
                        "bow":
                              {
                               ne11: frequency1,
                               ...
                               }
                        metadata1: "metadata1",
                        metadata2: "metadata2",
                        ...
                    }
        """

        # in order to load the vectorization in a mongodb database
        if mode in ['tf', 'onehot']:
            docterm = getattr(self, mode)(index=index, min_len=min_len,
                                          min_count=min_count, format_='docterm-matrix')

        elif mode == 'tfidf':
            docterm = getattr(self, mode)(index=index, normalize=normalize,
                                          min_len=min_len, min_count=min_count,
                                          format_='docterm-matrix')

        # elif mode == 'mi':
        else:
            docterm = getattr(self, mode)(index=index, min_len=min_len,
                                          min_count=min_count,
                                          format_='docterm-matrix')

        # keep the information
        if metadata is not None and len(metadata) > 0:
            for row in docterm:
                record = dict(row)
                record.update(metadata[row[self.group_by_field_key]])
                yield record
        else:
            for row in docterm:
                yield row



import abc
from typing import *

class LexiconMeta(metaclass=abc.ABCMeta):
    """
    Lexicon-based strategies are very efficient and simple methods. They make use of a sentiment _lexicon_
    to assign a polarity value to each text document by following a basic algorithm. A sentiment lexicon is a
    list of lexical features (e.g., words, phrase, etc.) which are labeled according to their semantic
    orientation (i.e. polarity) as either _positive_ (pos), _negative_ (neg), _neutral_ (obj).
    """

    def __init__(self, pos: Iterable[str] = None) -> None:
        """
        Args:
            pos (Tuple[str], optional, default=None): Some lexicons provide a polarity of a token for each part of speech that token may have in the speech. <br/> If not None, only the polarities for the parts of speech in the pos tuple will be considered.
        """

        pos_dict = {'n': 'NOUN', 'v': 'VERB', 'a': 'ADJ', 's': 'ADJ', 'r': 'ADV'}
        self.__PoS__ = [k for k, v in pos_dict.items() if pos is None or v in pos]

    @abc.abstractmethod
    def polarity(self, token:str) -> dict:
        """
        Returns the token polarities.

        Args:
            token (str): The text of the input token.

        Returns:
            dict: Polarities type as keys, sentiment values as values.

                    Example:
                    {
                     'pos' : 0.9,
                     'neg' : 0.1,
                     'obj' : 0.3
                    }
        """

    def polarities(self, bag_of_words: List[str]) -> dict:
        """
        Returns the polarities of the tokens in bag_of_words.

        Args:
            bag_of_words (List[str]): List of token.

        Returns:
            dict: Tokens as key, polarity dict as values.

                    Example:
                    {
                     token1: {
                                'pos' : 0.9,
                                'neg' : 0.1,
                                'obj' : 0.3
                              },
                     ...,
                     token2 : {...}
                     }
        """
        return {token: self.polarity(token) for token in bag_of_words}


from sklearn.cluster import KMeans as KM
from compling.unsupervisedLearning.unsupervised_learning import UnsupervisedLearning
from collections import defaultdict
from typing import *

class KMeans(UnsupervisedLearning):
    """
    Kmeans algorithm is an iterative algorithm that tries to partition the dataset into K pre-defined distinct non-overlapping subgroups (clusters) where each data point belongs to only one group.
    """

    def run(self, k:int, **kwargs:object) -> Dict[int, List[str]]:
        """
        Runs KMeans clustering.

        Args:
            k (int): The number of clusters to form as well as the number of centroids to generate.
            kwargs (object, optional): See https://scikit-learn.org/stable/modules/generated/sklearn.cluster.KMeans.html .

        Returns:
            Dict[int, List[str]]: Id of the clusters as keys, id of the vectors belonging in those cluster as values. <br/> The id of the cluster are automatically generated as progressive int.

                    Example:
                    >> k = KMeans(vectors, grouped_by_field="doc_id")
                    >> print(k.run(4))
                     {
                        0: [id_vector1, id_vector1, id_vector10, id_vector15, ...]
                        ...
                        3: [id_vector2, id_vector3, id_vector6, id_vector14, ...]
                     }
        """

        kwargs['n_clusters'] = k
        kmeans = KM(**kwargs)
        self.clusters = kmeans.fit_predict(self.data.T.values)

        cluster = defaultdict(list)

        data = {}
        for i, c in enumerate(self.clusters):
            cluster[c].append(self.data.T.index[i])
        for c, labels in cluster.items():
            data[str(c)] = labels

        return data


import csv
import pkg_resources
from collections import defaultdict
from compling.languages import italian
from compling.lexicons.lexiconMeta import LexiconMeta

csv.field_size_limit(1000000)

# Path to Sentix Lexicon: csv file
__sentix__ = pkg_resources.resource_filename('compling', 'senti-lexicons/sentix.csv')

# Parts of speech of the token we want to analyze
__pos__ = italian.pos_token_sentiment


class Sentix(LexiconMeta):
    """
    Sentiment Lexicon for Italian Language.

    http://valeriobasile.github.io/twita/sentix.html
    """

    def __init__(self):

        module = __import__('compling.languages.italian')
        language = getattr(getattr(module, 'languages'), 'italian')
        super().__init__(language.pos_token_sentiment)

        # lexicon object
        self.__sentiment__ = defaultdict(lambda: defaultdict(list))

        # read from file
        with open(__sentix__, encoding='utf-8', mode='r') as f:
            rows = csv.reader(f, delimiter='\t')

            columns = next(rows)

            for row in rows:
                record = dict(zip(columns, row))
                lemma, synset = record['lemma'].lower(), record['POS']
                self.__sentiment__[lemma][synset].append(record)

    def polarity(self, token: str) -> dict:
        """
        Returns a float for sentiment strength based on the input text.

        * The positive polarity of a token is calculated as the average of the positive polarities of its synsets that have a positive or negative polarity greater than 0.
        * The negative polarity of a token is calculated as the average of the positive polarities of its synsets that have a positive polarity or negative greater than 0.
        * The neutral polarity of a token is calculated as the average of the positive polarities of its synsets that have a neutral polarity greater than 0.

        Args:
            token (str): The text of the input token.

        Returns:
            dict: A positive, a negative and a neutral value.
        """

        # text processing
        is_negated = True if token.startswith('NOT_') else False
        if is_negated:
            token = token.replace("NOT_", "")
        token = token.lower()

        polarities = dict({'neg': 0.0, 'pos': 0.0, 'obj': 0.0})
        count = dict({'neg': 0, 'pos': 0, 'obj': 0})

        if token not in self.__sentiment__:
            return polarities

        for synset in self.__sentiment__[token]:
            if synset not in self.__PoS__:
                continue

            for record in self.__sentiment__[token][synset]:

                polarity = float(record['polarity'])
                pos = 0 if polarity < 0 else polarity
                neg = 0 if polarity > 0 else -polarity
                obj = 1 - float(record['intensity'])

                if is_negated:
                    pos, neg = neg, pos

                if pos > 0:
                    count['pos'] += 1
                if neg > 0:
                    count['neg'] += 1

                polarities['pos'] += pos
                polarities['neg'] += neg

                if obj > 0:
                    count['obj'] += 1

                polarities['obj'] += obj

        # Normalization
        if count['pos'] > 0 or count['neg'] > 0:
            polarities['pos'] /= max(count["pos"], count["neg"])
            polarities['neg'] /= max(count["pos"], count["neg"])
        if count['obj'] > 0:
            polarities['obj'] /= count['obj']

        return polarities


language = 'dutch'
iso639 = 'nl'

spacy_model_sm = iso639 + '_core_news_sm'
spacy_model_md = iso639 + '_core_news_md'
spacy_model_lg = iso639 + '_core_news_lg'

sentiment_lexicon_class = None
pos_token_sentiment = None

negations = ['zelfs niet', 'zelfs niet', 'zelfs niet', 'niemand', 'niemand', 'niets', 'niets', 'helemaal niet',
             'niet', 'minder', 'nooit meer', 'nooit', 'nee']


__all__ = ['sentiwordnet', 'vader']
from compling.analysis.sentiment.lexicon.impl.sentiwordnet import Sentiwordnet
from compling.analysis.sentiment.lexicon.impl.vader import Vader
from compling.analysis.sentiment.lexicon.impl.sentix import Sentix


language = 'polish'
iso639 = 'pl'

spacy_model_sm = iso639 + '_core_news_sm'
spacy_model_md = iso639 + '_core_news_md'
spacy_model_lg = iso639 + '_core_news_lg'

sentiment_lexicon_class = None
pos_token_sentiment = None

negations = ['nie', 'brak', 'ani', 'nigdy', 'nic', 'nieważność', 'nicość']


import csv
import pkg_resources
from compling.analysis.sentiment.lexicon.lexicon import Lexicon

csv.field_size_limit(1000000)

__fannix__ = pkg_resources.resource_filename('compling', 'senti-lexicons/fannix.csv')

class Fannix(Lexicon):
    """Sentix Lexicon Implementation of Lexicon Abastract Class.

    Chinese Language.

    https://github.com/fannix/Chinese-Sentiment-Lexicon/
    """

    def __init__(self) -> None:
        """
        **\_\_init\_\_**: Creates a new Sentiwordnet Chinese object.
        """

        self.sentiment = dict()

        with open(__fannix__, encoding='utf-8', mode='r') as f:
            rows = csv.reader(f, delimiter=',')

            columns = next(rows)

            for row in rows:
                record = dict(zip(columns, row))
                lemma = record['lemma'].lower()
                self.sentiment[lemma] = record


    def polarity(self, token: str) -> dict:
        """
        Returns the token polarities.

        Args:
            token (str): The text of the input token.

        Returns:
            dict: Polarities type as keys, sentiment values as values.

                    Example:
                    {
                     'pos' : 0.9,
                     'neg' : 0.1,
                     'obj' : 0.3
                    }
        """

        polarities = dict({'neg': 0.0, 'pos': 0.0, 'obj': 0.0})
        count = dict({'neg': 0, 'pos': 0, 'obj': 0})

        if token.replace("NOT_", "").lower() not in self.sentiment:
            return polarities

        sentiment = self.sentiment[token.replace("NOT_", "").lower()]

        pos = float(sentiment['pos'])
        neg = float(sentiment['neg'])
        obj = float(sentiment['obj'])

        if token.startswith('NOT_'):

            if neg > 0:
                count['pos'] += 1
            if pos > 0:
                count['neg'] += 1

            polarities['neg'] += pos
            polarities['pos'] += neg
        else:

            if pos > 0:
                count['pos'] += 1
            if neg > 0:
                count['neg'] += 1

            polarities['pos'] += pos
            polarities['neg'] += neg

        if obj > 0:
            count['obj'] +=1

        polarities['obj'] += obj

        for k in polarities.keys():
            if count[k] > 0:
                if k in ("pos", "neg"):
                    polarities[k] /= max(count["pos"], count["neg"])
                else:
                    polarities[k] /= count[k]

        return polarities



language = 'danish'
iso639 = 'da'

spacy_model_sm = iso639 + '_core_news_sm'
spacy_model_md = iso639 + '_core_news_md'
spacy_model_lg = iso639 + '_core_news_lg'

sentiment_lexicon_class = None
pos_token_sentiment = None

negations = ['ikke engang', 'ikke en gang', 'ingen', 'nogen', 'ingen', 'intet', 'slet ikke', 'ikke', 'mindre',
             'aldrig igen', 'aldrig', 'nej']


import inspect
import typing
from typing import get_type_hints
from inspect import getcallargs

def __get_class_that_defined_method__(meth):
    """Returns the class that defined method"""

    if inspect.ismethod(meth):
        for cls in inspect.getmro(meth.__self__.__class__):
           if cls.__dict__.get(meth.__name__) is meth:
                return cls
        meth = meth.__func__  # fallback to __qualname__ parsing

    if inspect.isfunction(meth):
        cls = getattr(inspect.getmodule(meth),
                      meth.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0])
        if isinstance(cls, type):
            return cls

    return getattr(meth, '__objclass__', None)  # handle special descriptor objects

def invokesuper(func):
    """method descriptor: invokes the super method."""

    def onCall(*args, **kwargs):
        _ = getattr(__get_class_that_defined_method__(func).__bases__[0], func.__name__)(*args, **kwargs)
        return func(*args, **kwargs)

    onCall.__name__ = func.__name__
    return onCall

def paramtypecheck(func):
    """method descriptor: check the parameter types."""

    def onCall(*args, **kwargs):
        hints = get_type_hints(func)

        call_args = getcallargs(func, *args, **kwargs)
        is_method = inspect.ismethod(func)

        # param index
        for param, paramtype in hints.items():
            if param == 'self' and is_method or param == 'return':
                continue

            # check if param is Iterable[str]
            if paramtype == typing.Iterable[str]:
                if not all([isinstance(s, str) for s in call_args[param]]):
                    raise TypeError("{} must be Iterable[str]".format(param))
                else:
                    continue

            # check if param is optional
            is_optional = not type(paramtype) == type
            paramtype = [paramtype] if not is_optional else paramtype.__args__

            if not is_optional:
                paramtypename = paramtype[0].__name__
            else:
                paramtypename = 'typing.Union[{}]'.format(", ".join(p.__name__ for p in paramtype))

            # Type Error
            if not type(call_args[param]) in paramtype:
                raise TypeError("{} must be {}, not {}".format(param, paramtypename,
                                                               type(call_args[param]).__name__))

        return func(*args, **kwargs)

    onCall.__name__ = func.__name__
    return onCall

def paramtypechecktarget(params):
    """method descriptor: check the parameter types."""

    def paramtypecheckt(func):
        """method descriptor: check the parameter types."""

        def onCall(*args, **kwargs):
            hints = get_type_hints(func)

            call_args = getcallargs(func, *args, **kwargs)
            is_method = inspect.ismethod(func)

            # param index
            for param, paramtype in hints.items():
                if param not in params:
                    continue

                # check if param is Iterable[str]
                if paramtype == typing.Iterable[str]:
                    if not all([isinstance(s, str) for s in call_args[param]]):
                        raise TypeError("{} must be Iterable[str]".format(param))
                    else:
                        continue

                # check if param is optional
                is_optional = not type(paramtype) == type
                paramtype = [paramtype] if not is_optional else paramtype.__args__

                if not is_optional:
                    paramtypename = paramtype[0].__name__
                else:
                    paramtypename = 'typing.Union[{}]'.format(", ".join(p.__name__ for p in paramtype))

                # Type Error
                if not type(call_args[param]) in paramtype:
                    raise TypeError("{} must be {}, not {}".format(param, paramtypename,
                                                                   type(call_args[param]).__name__))

            return func(*args, **kwargs)

        onCall.__name__ = func.__name__
        return onCall

    return paramtypecheckt

def f2att(function):
    """On first call, it stores the result of the method in an attribute. <br/>
       On next ones, the result will be returned without executing the method."""
    def onCall(self, *args):
        attribute = function.__name__ + '_'
        if not hasattr(self, attribute):
            self.__dict__[attribute] = function(self, *args)
        return self.__dict__[attribute]

    return onCall


import abc
from typing import *


class Paragraph(metaclass=abc.ABCMeta):
    """
    Represents a generic paragraph, which is a sequence of sentence in the input.
    A paragraph is identified by grouping a list of sentence.
    """

    @abc.abstractmethod
    def named_entities(self) -> List[str]:
        """Returns named entities in the paragraph."""

    @abc.abstractmethod
    def text(self) -> str:
        """Return the text of the paragraph."""

    @abc.abstractmethod
    def lower(self) -> str:
        """Return the lowercase text of the paragraph."""

    @abc.abstractmethod
    def lemma(self) -> str:
        """Returns a version of the paragraph text with tokens replace by their lemma."""

    @abc.abstractmethod
    def negations(self) -> str:
        """Return the text of the paragraph where the text of negated tokens are preceded by the prefix 'NOT_'."""

    @abc.abstractmethod
    def stem(self) -> str:
        """Returns a version of the paragraph text with tokens replace by their stem."""

    @abc.abstractmethod
    def add_metadata(self, metadata: dict) -> None:
        """Adds metadata as attributes of the paragraph."""

    @abc.abstractmethod
    def to_dict(self) -> dict:
        """Converts the paragraph to a dict record."""

    def sentiment(self) -> Dict[str, float]:
        """
        Returns the paragraph polarities. <br/>
        Each polarity value is between 0 and 1. They are calculated as the average of the sentences polarities greater than 0. <br/>
        _p_ in _[0,1]_

        Returns:
            Dict[str, float]: Polarities type as keys, sentiment values as values.
        """


import pkg_resources
from typing import *
from configparser import ConfigParser


class ConfigManager:

    def __init__(self) -> None:
        self.__config_path__ = pkg_resources.resource_filename('compling', 'config.ini')
        self.__config_parser__ = None
        self.__config_load__()

    def __config_load__(self) -> None:
        """Loads set-up parameters."""

        self.__config_parser__ = ConfigParser(interpolation=None)
        self.__config_parser__.read(self.__config_path__)

    def cat(self) -> None:
        """Shows set-up parameters."""
        with open(self.__config_path__, mode='r') as f:
            cat = f.read()
        print(cat)

    def updates(self, sections_key_values: List[Tuple[str]]) -> None:
        """Updates some set-up parameters.

        Args:
           sections_key_values (List[Tuple[str]]): Each tuple must consist of 3 strings: (section, key, value)
        """

        for section, key, value in sections_key_values:
            self.__config_parser__.set(section, key, value)

        # Write changes back to file
        with open(self.__config_path__, mode='w') as f_config:
            self.__config_parser__.write(f_config)

    def update(self, s: str, k: str, v: str) -> None:
        """Updates the set-up parameter "k" with the value "v" in the section "s".

        Args:
           s (str): Section name.
           k (str): Section key.
           v (str): New k value.
        """

        self.__config_parser__.set(s, k, v)

        # Write changes back to file
        with open(self.__config_path__, mode='w') as f_config:
            self.__config_parser__.write(f_config)

    def get(self, s: str, k: str) -> str:
        """Returns the value "v" for the key "k" in the section "s" """

        return self.__config_parser__[s][k]

    def get_section(self, s: str) -> dict:
        """Returns the set-up parameters in a section "s" """

        return dict(self.__config_parser__._sections[s])

    def reset(self) -> None:
        """Resets the set-up parameters."""

        with open(pkg_resources.resource_filename('compling', 'default_config.ini'), mode='r') as f:
            default = f.read()
        with open(self.__config_path__, mode='w') as f:
            f.write(default)

        self.__config_load__()

    def __whereisconfig__(self) -> str:
        """Shows the configuration file location."""

        return pkg_resources.resource_filename('compling', self.__config_path__)



from typing import *
from compling.config import ConfigManager
import compling.lexicalAnalysis.tokenization.paragraph_abc as abcpara
from compling.lexicalAnalysis.tokenization.impl.sentence import Sentence, f2att


class Paragraph(abcpara.Paragraph):
    def __init__(self, para_id: int, sent_list: List[Sentence], config: ConfigManager, metadata: dict) -> None:
        """
        Represents a generic paragraph, which is a sequence of sentence in the input.
        A paragraph is identified by grouping a list of sentence.

        Args:
           para_id (int): unique paragraph identifier.
           sent_list (List[Sentence]): The list of sentences making up the paragraph.
           config (ConfigManager): a ConfigManager object.
           metadata (dict): token metadata as a key-value pairs.

        Returns:
           None.
        """

        # unique identifiers
        self.para_id_ = para_id

        # paragraph
        self.text_ = " . ".join([sentence.text() for sentence in sent_list])

        # sentence list
        self.sent_list = sent_list

        # add some field only if the value in config.ini file is True
        for k, v in config.get_section(s='Paragraph_record').items():
            if int(v):
                setattr(self, k + '_', getattr(self, k)())

        # add metadata for this paragraph record
        self.add_metadata(metadata)

    @f2att
    def named_entities(self) -> List[Dict[str, Union[str, int]]]:
        """Returns named entities in the paragraph."""

        named_entities = list()
        for i, sent in enumerate(self.sent_list):
            offset = sent.token_list[0].para_pos_

            for ne in sent.named_entities():
                ne_ = dict()
                ne_['para_pos_start_'] = ne['sent_pos_start_'] + offset
                ne_['para_pos_end_'] = ne['sent_pos_end_'] + offset
                ne_['text_'] = ne['text_']
                named_entities.append(ne_)

        return named_entities

    def lang(self) -> str:
        """Returns the language of the paragraph."""
        return self.sent_list[0].lang()

    def text(self) -> str:
        """Return the text of the paragraph."""
        return self.text_

    def lower(self) -> str:
        """Return the lowercase text of the paragraph."""
        return self.text_.lower()

    @f2att
    def lemma(self) -> str:
        """Returns a version of the paragraph text with tokens replace by their lemma."""

        return " . ".join([sentence.lemma() for sentence in self.sent_list])

    @f2att
    def negations(self) -> str:
        """Return the text of the paragraph where the text of negated tokens are preceded by the prefix 'NOT_'."""

        return " . ".join([sent.negations() for sent in self.sent_list])

    @f2att
    def sentiment(self) -> Dict[str, float]:
        """
        Returns the paragraph polarities. <br/>
        Each polarity value is between 0 and 1. They are calculated as the average of the sentences polarities greater than 0. <br/>
        _p_ in _[0,1]_

        Returns:
            Dict[str, float]: Polarities type as keys, sentiment values as values.
        """

        sentiment_values = {'pos': 0.0, 'neg': 0.0}
        count_gt0 = {'pos': 0, 'neg': 0}

        for sent in self.sent_list:
            for k in sentiment_values.keys():
                if sent.sentiment()[k] > 0:
                    count_gt0[k] += 1
                sentiment_values[k] += sent.sentiment()[k]

        for k in sentiment_values.keys():
            if count_gt0[k] > 0:
                if k in ("pos", "neg"):
                    sentiment_values[k] /= max(count_gt0["pos"], count_gt0["neg"])
                else:
                    sentiment_values[k] /= count_gt0[k]

        return sentiment_values

    @f2att
    def stem(self) -> str:
        """Returns a version of the paragraph text with tokens replace by their stem."""
        return " . ".join([sentence.stem() for sentence in self.sent_list])

    def add_metadata(self, metadata: dict) -> None:
        """Adds metadata as attributes of the paragraph."""
        for k, v in metadata.items():
            if hasattr(self, k + '_'):
                continue
            setattr(self, k + '_', v)

    def to_dict(self) -> dict:
        """Converts the paragraph to a dict record."""
        return {k: v for k, v in self.__dict__.items() if k.endswith('_')}



from typing import *
from compling.config import ConfigManager
import compling.lexicalAnalysis.tokenization.document_abc as abcdoc
from compling.lexicalAnalysis.tokenization.impl.paragraph import Paragraph
from compling.descriptor import f2att


class Document(abcdoc.Document):
    def __init__(self, doc_id: int, para_list: List[Paragraph], config: ConfigManager, metadata: dict) -> None:
        """
        Represents a generic document, which is a sequence of paragraph in the input.
        A document is identified by grouping a list of paragraph.

        Args:
           doc_id (int): unique document identifier.
           para_list (List[Paragraph]): The list of paragraphs making up the document.
           config (ConfigManager): a ConfigManager object.
           metadata (dict): token metadata as a key-value pairs.

        Returns:
           None.
       """

        # unique identifiers
        self.doc_id_ = doc_id

        # document
        self.text_ = " . ".join([sentence.text() for sentence in para_list])

        # paragraph list
        self.para_list = para_list

        # add some field only if the value in config.ini file is True
        for k, v in config.get_section(s='Document_record').items():
            if int(v):
                setattr(self, k + '_', getattr(self, k)())

        # add metaddata for this document record
        self.add_metadata(metadata)

    def lang(self) -> str:
        """Returns the language of the document."""
        return self.para_list[0].lang()

    @f2att
    def sentiment(self) -> Dict[str, float]:
        """
        Returns the document polarities. <br/>
        Each polarity value is between 0 and 1. They are calculated as the average of the paragraphs polarities greater than 0. <br/>
        _p_ in _[0,1]_

        Returns:
            Dict[str, float]: Polarities type as keys, sentiment values as values.
        """


        sentiment_values = {'pos': 0.0, 'neg': 0.0}
        count_gt0 = {'pos': 0, 'neg': 0}

        for para in self.para_list:
            for k in sentiment_values.keys():
                if para.sentiment()[k] > 0:
                    count_gt0[k] += 1
                sentiment_values[k] += para.sentiment()[k]

        for k in sentiment_values.keys():
            if count_gt0[k] > 0:
                if k in ("pos", "neg"):
                    sentiment_values[k] /= max(count_gt0["pos"], count_gt0["neg"])
                else:
                    sentiment_values[k] /= count_gt0[k]

        return sentiment_values

    @f2att
    def named_entities(self) -> List[Dict[str, Union[str, int]]]:
        """Returns named entities in the document."""

        named_entities = list()
        for i, para in enumerate(self.para_list):
            for j, sent in enumerate(para.sent_list):
                offset = para.sent_list[j].token_list[0].doc_pos_

                for ne in sent.named_entities():
                    ne_ = dict()
                    ne_['doc_pos_start_'] = ne['sent_pos_start_'] + offset
                    ne_['doc_pos_end_'] = ne['sent_pos_end_'] + offset
                    ne_['text_'] = ne['text_']
                    named_entities.append(ne_)

        return named_entities

    def text(self) -> str:
        """Returns the text of the document."""
        return self.text_

    def lower(self) -> str:
        """Returns the lower of the document."""
        return self.text_.lower()

    @f2att
    def lemma(self) -> str:
        """Returns a version of the document text with tokens replace by their lemma."""
        return " . ".join([sentence.lemma() for sentence in self.para_list])

    @f2att
    def negations(self) -> str:
        """Returns the text of the document where the text of negated tokens are preceded by the prefix 'NOT_'."""

        return " . ".join([para.negations() for para in self.para_list])

    @f2att
    def stem(self) -> str:
        """Returns a version of the document text with tokens replace by their stem."""
        return " . ".join([sentence.stem() for sentence in self.para_list])

    def add_metadata(self, metadata: dict) -> None:
        """Adds metadata as attributes of the document."""
        for k, v in metadata.items():
            if hasattr(self, k + '_'):
                continue
            setattr(self, k + '_', v)

    def to_dict(self) -> dict:
        """Converts the document to a dict record."""
        return {k: v for k, v in self.__dict__.items() if k.endswith('_')}



from compling.lexicons.lexiconMeta import LexiconMeta
from vaderSentiment.vaderSentiment import SentimentIntensityAnalyzer


class Vader(LexiconMeta):
    """VADER Lexicon for English Language."""

    def __init__(self) -> None:

        super().__init__()
        self.__sentiment__ = SentimentIntensityAnalyzer()

    def polarity(self, token: str) -> dict:
        """
        Returns a float for sentiment strength based on the input text.
        Positive values are positive valence, negative value are negative valence, neutral value are neutral valence.'

        Args:
            token (str): The input token.

        Returns:
            dict: A positive, a negative and a neutral value.
        """

        # text processing
        is_negated = True if token.startswith('NOT_') else False
        if is_negated:
            token = token.replace("NOT_", "")
        token = token.lower()

        polarities = self.__sentiment__.polarity_scores(token)

        # polarity inversion
        if is_negated:
            polarities['neg'], polarities['pos'] = polarities['pos'], polarities['neg']

        polarities['obj'] = polarities['neu']
        del polarities['neu']
        del polarities['compound']

        return polarities


from compling.lexicalAnalysis.vectorization import vectorization
from tqdm import tqdm
import copy
import math
from typing import *
from collections import defaultdict


class Vectorizer(vectorization.Vectorization):
    """
    The process of converting text into vector is called vectorization.
    The set of corpus documents vectorized corpus makes up the Vector Space Model, which can have a sparse or
    dense representation.

    A Vectorizer object allows you to create vectors grouping tokens for arbitrary fields.

    Grouping tokens by:

       * _doc_id_: you 're creating document vectors;
       * _sent_id_: you 're creating sentence vectors;
       * _author_: you're creating author vectors (each token must have an 'author' field);
       ...

    You can group by multiple fields.

    You can also choose the text field the tokens will be grouped by too. (e.g. lemma, text, stem, ...)

    It offers several functions to set the vector components values, such as:

       * **One-hot encoding** (binary values 0/1);
       * **Tf** (Term Frequency);
       * **TfIdf** (Term Frequency Inverse Document Frequency);
       * **MI** (Mutual Information)

    You can specify the vectorization representation format: Term x Document matrix, Postings list.
    """

    def postings2docterm(self, postings: Dict[str, dict]) -> Iterable[dict]:
        """
        Converts a postings list to docterm matrix.

        Args:
            postings (Dict[str, dict]): Postings list.

                    Example:
                    >> print(postings)
                    {
                     "token1":
                           {
                             "doc1": frequency1,
                             ...
                           },
                     ...
                     }

        Returns:
            Iterable[dict]: Docterm matrix.

                    Example:
                    >> print(next(postings2docterm(postings)))
                    {
                     "doc": "doc1",
                     "bow":
                           {
                            token1: frequency1,
                            ...
                            }
                     }
        """

        docterm = defaultdict(dict)
        for token in postings:
            for doc, relevance in postings[token].items():
                docterm[doc].update({token: relevance})

        for doc in docterm:
            yield {self.group_by_field_key: doc, 'bow': docterm[doc]}

    def docterm2postings(self, docterm: Iterable[dict]) -> Dict[str, dict]:
        """
        Converts a docterm matrix to a postings list.

        Args:
            docterm (Iterator[dict]): Docterm matrix.

                    Example:
                    >> print(next(docterm))
                    {
                     "doc": "doc1",
                     "bow":
                           {
                            token1: frequency1,
                            ...
                            }
                     }
                     # bow: Bag of Words
        Returns:
            Dict[str, dict]: Postings list.

                    Example:
                    >> print(docterm2postings(docterm))
                    {
                     "token1":
                           {
                             "doc1": frequency1,
                             ...
                           },
                     ...
                     }
        """

        postings = defaultdict(dict)
        for record in docterm:
            for token, relevance in record['bow'].items():
                postings[token].update({record[self.group_by_field_key]: relevance})

        return postings

    def tf(self, tokens: Iterable[dict], pos: Tuple[str, ...] = ("NOUN", "VERB", "PROPN", "NGRAM"),
           dep: Tuple[str, ...] = None, lang: str = None, min_len: int = 0,
           min_count: int = 0, format_: str = "docterm-matrix", boolean=False) -> Union[Dict[str, dict], Iterable[dict]]:
        """
        Calculates Term Frequency.

        Args:
            tokens (Iterable[dict]): Stream of token records produced during the Tokenization stage.

                    Example:
                    >>> token = next(tokens)
                    >>> print(token)
                    {
                    "token_id_": 0,
                    "sent_id_": 0,
                    "para_id_": 0,
                    "doc_id_": 0,
                    "text_": "When",
                    "sent_pos_": 0,
                    "para_pos_": 0,
                    "doc_pos_": 0,
                    "lower_": "when",
                    ...
                    }

            pos (Tuple[str, ...], optional, default=NOUN,VERB,PROPN,ADJ): Part of speech of the tokens. <br/> Filters the most informative tokens: tokens whose PoS tag is in the 'pos' list. <br/> If None, the filter will be ignored.
            dep (Tuple[str, ...], optional, default=None): Dependencies of the tokens. <br/> Filters the most informative tokens: tokens whose dep tag is in the 'dep' list. <br/> If None, the filter will be ignored.
            lang (str, optional, default=None): Filters the tokens of the language (iso639). <br/> If None, the filter will be ignored.
            min_len (int, optional, default=0): The minimum length a token must have to be considered.
            min_count (int, optional, default=0): The minimum frequency a token must have in the corpus to be considered.
            format_ (str, optional, default=docterm-matrix): Format of the vectorization records.

                Valid input values:
                ["postings-list", "docterm-matrix"]

            boolean (bool, optional, default=False): If True return Boolean Vectors else Term Frequency. <br/> If True, the result is the same as onehot method.

        Returns:
            Union[Dict[str, dict], Iterable[dict]]: If format_ == 'postings-list', returns a postings list; else a docterm matrix.

                    Example:
                    [format_ == "postings-list"]

                    {
                        "token1":
                              {
                                "doc1": tf1,
                                ...
                              },
                        ...
                    }

                     Example:
                    [format_ == "docterm-matrix"]

                    [{
                        "doc": "doc1",
                        "bow":
                              {
                               token1: tf1,
                               ...
                               }
                    },
                    ...
                    ]
        """

        if not hasattr(self, 'tf_postings'):
            filters = dict()
            for k, v in self.nlp.config.get_section(s='Vector_filter').items():
                if v != 'None':
                    filters[k] = bool(int(v))

            def __filter__(__token__):
                for key, value in {'dep_': dep, 'pos_': pos}.items():
                    if value is not None and len(value) != 0 and __token__[key] not in value:
                        return False

                if len(__token__[self.token_field]) < min_len:
                    return False

                for key, value in filters.items():
                    if not __token__[key + '_'] == value:
                        return False

                if lang is not None and lang != __token__['lang_']:
                    return False

                return True

            postings = defaultdict(lambda: defaultdict(int))
            for token in tqdm(tokens, desc="TF calculation in progress...", position=0, leave=True):
                if not __filter__(token):
                    continue

                key = " + ".join([token[g] for g in self.group_by_field]).replace('.', '\u002E')

                # replace . (ngrams like "commissione_._signore")
                token_field = token[self.token_field].replace('.', '\u002E')
                if boolean:
                    postings[token_field][key] = 1
                else:
                    postings[token_field][key] += 1

            for token in copy.deepcopy(postings):
                if sum(postings[token].values()) < min_count:
                    del postings[token]

            self.tf_postings = postings

        if format_ == 'postings-list':
            return self.tf_postings
        else:
            return self.postings2docterm(self.tf_postings)

    def reset_tf(self):
        """If present, removes the tf postings list from memory.
         It will be re-calculated on the next run call."""
        if hasattr(self, 'tf_postings'):
            delattr(self, 'tf_postings')

    def tfidf(self, tokens: Iterable[dict], normalize=True,
              pos: Tuple[str, ...] = ("NOUN", "VERB", "PROPN", "NGRAM"),
              dep: Tuple[str, ...] = None, lang: str = None, min_len: int = 0,
              min_count: int = 0, format_: str = 'docterm-matrix') -> Union[Dict[str, dict], Iterable[dict]]:
        """
        Calculates Term Frequency Inverse Document Frequency.


        Args:
            tokens (Iterable[dict]): Stream of token records produced during the Tokenization stage.

                    Example:
                    >>> token = next(tokens)
                    >>> print(token)
                    {
                    "token_id_": 0,
                    "sent_id_": 0,
                    "para_id_": 0,
                    "doc_id_": 0,
                    "text_": "When",
                    "sent_pos_": 0,
                    "para_pos_": 0,
                    "doc_pos_": 0,
                    "lower_": "when",
                    ...
                    }

            normalize (bool, optional, default=True): If True, normalizes TF for max(TF).
            pos (Tuple[str, ...], optional, default=NOUN,VERB,PROPN,ADJ): Part of speech of the tokens. <br/> Filters the most informative tokens: tokens whose PoS tag is in the 'pos' list. <br/> If None, the filter will be ignored.
            dep (Tuple[str, ...], optional, default=None): Dependencies of the tokens. <br/> Filters the most informative tokens: tokens whose dep tag is in the 'dep' list. <br/> If None, the filter will be ignored.
            lang (str, optional, default=None): Filters the tokens of the language (iso639). <br/> If None, the filter will be ignored.
            min_len (int, optional, default=0): The minimum length a token must have to be considered.
            min_count (int, optional, default=0): The minimum frequency a token must have in the corpus to be considered.
            format_ (str, optional, default=docterm-matrix): Format of the vectorization records.

                    Valid input values:
                    ["postings-list", "docterm-matrix"]

        Returns:
            Union[Dict[str, dict], Iterable[dict]]: If format_ == 'postings-list', returns a postings list; else a docterm matrix.

                    Example:
                    [format_ == "postings-list"]

                    {
                        "token1":
                              {
                                "group_by_field1": tfid1,
                                ...
                              },
                        ...
                    }

                     Example:
                    [format_ == "docterm-matrix"]

                    [{
                        group_by_field: "group_by_field1",
                        "bow":
                              {
                               token1: tfid1,
                               ...
                               }
                    },
                    ...
                    ]
        """


        postings = self.tf(tokens, pos, dep, lang, min_len, min_count, format_='postings-list')

        docterm = self.postings2docterm(postings)
        max_tf = {}
        n = 0
        for doc in docterm:
            n += 1
            max_tf[doc[self.group_by_field_key]] = max(doc['bow'].values())

        # Corpus size: the number of documents in the corpus
        N = n

        # The tf_dict will be update to tfidf_dict
        tfidf_postings = copy.deepcopy(postings)

        bar = tqdm(total=len(tfidf_postings), desc="TFIDF calculation in progress... ", position=0, leave=True)

        # For each token in the postings list
        for token in tfidf_postings.keys():

            docs = tfidf_postings[token]

            # For each doc which the token occurs in
            for doc, tf in docs.items():
                if normalize:
                    tf = tf / max_tf[doc]

                # Document Frequency
                df = len(postings[token])

                # Inverse Document Frequency
                idf = math.log(N / df)

                # Term Frequency Inverse Document Frequency
                tfidf_postings[token].update({doc: tf * idf})
            bar.update(1)
        bar.close()

        if format_ == 'postings-list':
            return tfidf_postings
        else:
            return self.postings2docterm(tfidf_postings)

    def mi(self, tokens: Iterable[dict], pos: Tuple[str, ...] = ("NOUN", "VERB", "PROPN", "NGRAM"),
           dep: Tuple[str, ...] = None, lang: str = None, min_len: int = 0,
           min_count: int = 0, format_: str = 'docterm-matrix') -> Union[Dict[str, dict], Iterator[dict]]:
        """
        Calculates Mutual Information.

        Args:
            tokens (Iterable[dict]): Stream of token records produced during the Tokenization stage.

                    Example:
                    >>> token = next(tokens)
                    >>> print(token)
                    {
                    "token_id_": 0,
                    "sent_id_": 0,
                    "para_id_": 0,
                    "doc_id_": 0,
                    "text_": "When",
                    "sent_pos_": 0,
                    "para_pos_": 0,
                    "doc_pos_": 0,
                    "lower_": "when",
                    ...
                    }

            pos (Tuple[str, ...], optional, default=NOUN,VERB,PROPN,ADJ): Part of speech of the tokens. <br/> Filters the most informative tokens: tokens whose PoS tag is in the 'pos' list. <br/> If None, the filter will be ignored.
            dep (Tuple[str, ...], optional, default=None): Dependencies of the tokens. <br/> Filters the most informative tokens: tokens whose dep tag is in the 'dep' list. <br/> If None, the filter will be ignored.
            lang (str, optional, default=None): Filters the tokens of the language (iso639). <br/> If None, the filter will be ignored.
            min_len (int, optional, default=0): The minimum length a token must have to be considered.
            min_count (int, optional, default=0): The minimum frequency a token must have in the corpus to be considered.
            format_ (str, optional, default=docterm-matrix): Format of the vectorization records.

                    Valid input values:
                    ["postings-list", "docterm-matrix"]

        Returns:
            Union[Dict[str, dict], Iterable[dict]]: If format_ == 'postings-list', returns a postings list; else a docterm matrix.

                    Example:
                    [format_ == "postings-list"]

                    {
                        "token1":
                              {
                                "doc1": mi1,
                                ...
                              },
                        ...
                    }

                    Example:
                    [format_ == "docterm-matrix"]

                    [{
                        "doc": "doc1",
                        "bow":
                              {
                               token1: mi1,
                               ...
                               }
                    },
                    ...
                    ]
        """

        postings = self.tf(tokens, pos, dep, lang, min_len, min_count, format_='postings-list')
        docterm = {record[self.group_by_field_key]: record for record in self.postings2docterm(postings)}

        # x: group_by_field label
        # y: token
        n_tokens = sum([sum(postings[token].values()) for token in postings])

        mi_postings = defaultdict(lambda :defaultdict(int))
        for token in tqdm(postings, desc="Mutual Information calculation in progress...", position=0, leave=True):
            posting = defaultdict(int, postings[token])
            for label in posting:
                row = docterm[label]['bow']
                len_x = sum(posting.values())
                len_y = sum(row.values())

                p_x = len_x/n_tokens
                p_y = len_y/n_tokens
                p_x_y = posting[label]/n_tokens

                mi_postings[token][label] = p_x_y * math.log(p_x_y/(p_x*p_y))

        if format_ == 'postings-list':
            return mi_postings
        else:
            return self.postings2docterm(mi_postings)

    def onehot(self, tokens: Iterable[dict], pos: Tuple[str, ...] = ("NOUN", "VERB", "PROPN", "NGRAM"),
               dep: Tuple[str, ...] = None, lang: str = None, min_len: int = 0,
               min_count: int = 0, format_: str = 'docterm-matrix') -> Union[Dict[str, dict], Iterable[dict]]:
        """
        Calculates One-hot encoding.

        Args:
            tokens (Iterable[dict]): Stream of token records produced during the Tokenization stage.

                    Example:
                    >>> token = next(tokens)
                    >>> print(token)
                    {
                    "token_id_": 0,
                    "sent_id_": 0,
                    "para_id_": 0,
                    "doc_id_": 0,
                    "text_": "When",
                    "sent_pos_": 0,
                    "para_pos_": 0,
                    "doc_pos_": 0,
                    "lower_": "when",
                    ...
                    }

            pos (Tuple[str, ...], optional, default=NOUN,VERB,PROPN,ADJ): Part of speech of the tokens. <br/> Filters the most informative tokens: tokens whose PoS tag is in the 'pos' list. <br/> If None, the filter will be ignored.
            dep (Tuple[str, ...], optional, default=None): Dependencies of the tokens. <br/> Filters the most informative tokens: tokens whose dep tag is in the 'dep' list. <br/> If None, the filter will be ignored.
            lang (str, optional, default=None): Filters the tokens of the language (iso639). <br/> If None, the filter will be ignored.
            min_len (int, optional, default=0): The minimum length a token must have to be considered.
            min_count (int, optional, default=0): The minimum frequency a token must have in the corpus to be considered.
            format_ (str, optional, default=docterm-matrix): Format of the vectorization records.

                    Valid input values:
                    ["postings-list", "docterm-matrix"]

        Returns:
            Union[Dict[str, dict], Iterable[dict]]: If format_ == 'postings-list', returns a postings list; else a docterm matrix.

                    Example:
                    [format_ == "postings-list"]

                    {
                        "token1":
                              {
                                "doc1": 1,
                                "doc2": 0,
                                ...
                              },
                        ...
                    }

                    Example:
                    [format_ == "docterm-matrix"]

                    [{
                        "doc": "doc1",
                        "bow":
                              {
                               token1: 1,
                               ...
                               }
                    },
                    ...
                    ]
        """
        return self.tf(tokens, pos, dep, lang, min_len, min_count, boolean=True, format_=format_)

    def run(self, mode: str, tokens: Iterable[dict], normalize=True,
            pos: Tuple[str, ...] = ("NOUN", "VERB", "PROPN", "NGRAM"),
            dep: Tuple[str, ...] = None,
            lang: str = None, min_len=0, min_count=0,
            metadata: Dict[str, dict] = None) -> Iterable[dict]:
        """
        Runs the vectorization grouping token by self.token_field and self.group_by_field.

        Args:
            mode (str): Vectorization mode. <br/>

                    Valid input values:
                    ["tf", "tfidf", "mi", "onehot"]

            tokens (Iterable[dict]): Stream of token records produced during the Tokenization stage.

                    Example:
                    >>> token = next(tokens)
                    >>> print(token)
                    {
                    "token_id_": 0,
                    "sent_id_": 0,
                    "para_id_": 0,
                    "doc_id_": 0,
                    "text_": "When",
                    "sent_pos_": 0,
                    "para_pos_": 0,
                    "doc_pos_": 0,
                    "lower_": "when",
                    ...
                    }

            normalize (bool, optional, default=True): If True, normalizes TF for max(TF). <br/> This parameter is ignored if mode in ['tf', 'onehot', 'mi'].
            pos (Tuple[str, ...], optional, default=NOUN,VERB,PROPN,ADJ): Part of speech of the tokens. <br/> Filters the most informative tokens: tokens whose PoS tag is in the 'pos' list. <br/> If None, the filter will be ignored.
            dep (Tuple[str, ...], optional, default=None): Dependencies of the tokens. <br/> Filters the most informative tokens: tokens whose dep tag is in the 'dep' list. <br/> If None, the filter will be ignored.
            lang (str, optional, default=None): Filters the tokens of the language (iso639). <br/> If None, the filter will be ignored.
            min_len (int, optional, default=0): The minimum length a token must have to be considered.
            min_count (int, optional, default=0): The minimum frequency a token must have in the corpus to be considered.
            metadata (Dict[str, dict], optional, default=None): A dict containing metadata for each different group_by_field value.  <br/> Each vector will be associated with its own metadata.

                    Example:

                    {
                     "group_by_field1": {
                                         metadata1: 'metadata1',
                                         metadata2: 'metadata2',
                                         ...
                                        },
                     "group_by_field2": {
                                          metadata1: 'metadata11',
                                          metadata2: 'metadata22',
                                         ...
                                        },
                    ...
                    }


        Returns:
            Iterable[dict]: Returns a docterm matrix.

                    Example:
                    >> v = Vectorizer('doc_id', 'text')
                    >> record = v.run('tfidf', tokens)
                    >> print(next(record))
                    {
                        "doc": "doc1",
                        "bow":
                              {
                               token1: frequency1,
                               ...
                               }
                        metadata1: "metadata1",
                        metadata2: "metadata2",
                        ...
                    }
        """

        # in order to load the vectorization in a mongodb database
        if mode in ['tf', 'onehot']:
            docterm = getattr(self, mode)(tokens=tokens, pos=pos, dep=dep, lang=lang, min_len=min_len,
                                          min_count=min_count, format_='docterm-matrix')

        elif mode == 'tfidf':
            docterm = getattr(self, mode)(tokens=tokens, normalize=normalize, pos=pos, dep=dep, lang=lang,
                                          min_len=min_len, min_count=min_count,
                                          format_='docterm-matrix')

        # elif mode == 'mi':
        else:
            docterm = getattr(self, mode)(tokens=tokens, pos=pos, dep=dep, lang=lang, min_len=min_len,
                                          min_count=min_count,
                                          format_='docterm-matrix')

        # keep the information
        if metadata is not None and len(metadata) > 0:
            for row in docterm:
                record = dict(row)
                record.update(metadata[row[self.group_by_field_key]])
                yield record
        else:
            for row in docterm:
                yield row



from nltk.corpus import sentiwordnet as swn
from compling.analysis.sentiment.lexicon.lexicon import Lexicon
from typing import Tuple

class Sentiwordnet(Lexicon):
    """Sentiwordnet Lexicon Implementation of Lexicon Abastract Class."""

    def __init__(self, pos: Tuple[str, str] = None) -> None:
        """
        **\_\_init\_\_**: Creates a new Sentiwordnet object.

        Args:
            pos (Tuple[str], optional, default=None): SentiWordNet Lexicon returns a token polarities for each part of speech that token may have in the speech. <br/> If not None, filters the token polarities for part of speech in the pos tuple.

        This class sums the polarity value of the parts of speech you choose
        """

        pos_dict = {'n': 'NOUN', 'v': 'VERB', 'a': 'ADJ', 's': 'ADJ', 'r': 'ADV'}
        self.pos = [k for k, v in pos_dict.items() if pos is None or v in pos or len(pos) == 0]

    def polarity(self, token: str) -> dict:
        """
        Returns the token polarities.

        Args:
            token (str): The text of the input token.

        Args:
            dict: Polarities type as keys, sentiment values as values.

                    Example:
                    {
                     'pos' : 0.9,
                     'neg' : 0.1,
                     'obj' : 0.3
                    }
        """

        senti_synsets = [senti_synset for senti_synset in list(swn.senti_synsets(token.replace("NOT_", "")))
                 if senti_synset.synset._pos in self.pos]

        polarities = dict({'neg': 0.0, 'pos': 0.0, 'obj': 0.0})
        count = dict({'neg': 0, 'pos': 0, 'obj': 0})

        for senti_synset in senti_synsets:
            pos = senti_synset.pos_score()
            neg = senti_synset.neg_score()
            obj = senti_synset.obj_score()

            if token.startswith('NOT_'):

                if pos > 0:
                    count['neg'] += 1
                if neg > 0:
                    count['pos'] += 1

                polarities['neg'] += pos
                polarities['pos'] += neg
            else:

                if pos > 0:
                    count['pos'] += 1
                if neg > 0:
                    count['neg'] += 1

                polarities['pos'] += pos
                polarities['neg'] += neg

            if obj > 0:
                count['obj'] += 1
            polarities['obj'] += obj

        for p in ['pos', 'neg', 'obj']:
            if count[p] > 0:
                if p in ("pos", "neg"):
                    polarities[p] /= max(count["pos"], count["neg"])
                else:
                    polarities[p] /= count[p]

        return polarities




import abc
from typing import *

class Sentence(metaclass=abc.ABCMeta):
    """
    Represents a generic sentence, which is a sequence of tokens in the input.
    A sentence is identified using a tokenizer, which breaks up the sequence of characters in the input into a sequence
    of tokens.
    """

    @abc.abstractmethod
    def named_entities(self) -> List[str]:
        """Returns named entities in the sentence."""

    @abc.abstractmethod
    def text(self) -> str:
        """Return the text of the sentence."""

    @abc.abstractmethod
    def lower(self) -> str:
        """Return the lowercase version of the sentence text."""

    @abc.abstractmethod
    def negations(self) -> str:
        """Return the text of the sentence where the text of negated tokens are preceded by the prefix 'NOT_'"""

    @abc.abstractmethod
    def lemma(self) -> str:
        """Returns a version of the sentence text with tokens replace by their lemma."""

    @abc.abstractmethod
    def stem(self) -> str:
        """Returns a version of the sentence text with tokens replace by their stem."""

    @abc.abstractmethod
    def add_metadata(self, metadata: dict) -> None:
        """Adds metadata as attributes of the sentence."""

    @abc.abstractmethod
    def to_dict(self) -> dict:
        """Converts the sentence to a dict record."""

    @abc.abstractmethod
    def sentiment(self) -> Dict[str, float]:
        """
        Returns the sentence polarities. <br/>
        Each polarity value is between 0 and 1. They are calculated as the average of the tokens polarities greater than 0. <br/>
        _p_ in _[0,1]_

        Returns:
            Dict[str, float]: Polarities type as keys, sentiment values as values.

                    Example:
                    sentence.text = "damn your friend, I hate her"

                    >> sentence.sentiment()
                    {
                     'pos' : 0.125,
                     'neg' : 0.775,
                     'obj' : 0.100
                    }
        """


__all__ = ['fannix', 'sentiwordnet', 'vader', 'sentix']

from compling.lexicons.vader import Vader
from compling.lexicons.fannix import Fannix
from compling.lexicons.sentix import Sentix
from compling.lexicons.sentiwordnet import Sentiwordnet


import abc
from compling.embeddings.embeddings import Embeddings
from typing import *


class WordEmbeddings(Embeddings, metaclass=abc.ABCMeta):
    """Word Embedding abstract class."""

    def most_similar_threshold(self, positive: Union[list, str] = None,
                               negative: Union[list, str] = None,
                               threshold: float = 0.75) -> Dict[str, float]:
        """
        Finds the most similar vectors, whose cosine similarity is lower than threshold.
        Positive words/doc_ids contribute positively towards the similarity, negative words/doc_ids negatively.
        This method computes cosine similarity between a simple mean of the projection weight vectors of the
        given words/doc_ids and the vectors for each word/doc_id in the model.

        Args:
            positive (Union[list, str]): List of words/doc_ids that contribute positively.
            negative (Union[list, str]): List of words/doc_ids that contribute negatively.
            threshold (float, optional, default=0.75): Cosine similarity threshold.

        Returns:
            Dict[str, float]: Words/doc_ids as keys, cosine similarity values as values.
        """

        vectors = self.model.wv
        n = len(vectors.vocab)
        mst_res = {item[0]: item[1] for item in vectors.most_similar(positive=positive, negative=negative, topn=n)
                   if item[1] > threshold}
        return mst_res

    def most_similar(self, positive: Union[list, str] = None,
                     negative: Union[list, str] = None,
                     topn: int = 10) -> Dict[str, float]:
        """
        Finds the top-N most similar words/doc_ids.
        Positive words/doc_ids contribute positively towards the similarity, negative words/doc_ids negatively.
        This method computes cosine similarity between a simple mean of the projection weight vectors of the
        given words/doc_ids and the vectors for each word/doc_id in the model.

        Args:
            positive (Union[list, str]): List of words/doc_ids that contribute positively.
            negative (Union[list, str]): List of words/doc_ids that contribute negatively.
            topn (int, optional, default=10): Number of top-N similar words to return.

        Returns:
            Dict[str, float]: Words/doc_ids as keys, cosine similarity values as values.
        """
        vectors = self.model.wv
        return dict(vectors.most_similar(positive=positive, negative=negative, topn=topn))

    def similarity(self, word1: str, word2: str) -> float:
        """
        Computes cosine similarity between two words/doc_ids.

        Args:
            id1 (str): First word/doc_id.
            id2 (str): First word/doc_id.

        Returns:
            float: Cosine similarity between two vectors.
        """

        vectors = self.model.wv
        return vectors.similarity(word1, word2)

    def run(self, size=100, sg=0, alpha=0.025, window=5, min_count=5, workers=3,
            min_alpha=0.001, epochs=25) -> None:
        """
        Trains a Neural Network and builds a Word Embeddings Model.

        See: https://radimrehurek.com/gensim/models/word2vec.html ."""

        super(WordEmbeddings, self).run(size=size, sg=sg, alpha=alpha, window=window, min_count=min_count,
                                        workers=workers, min_alpha=min_alpha, epochs=epochs)



language = 'german'
iso639 = 'de'

spacy_model_sm = iso639 + '_core_news_sm'
spacy_model_md = iso639 + '_core_news_md'
spacy_model_lg = iso639 + '_core_news_lg'

sentiment_lexicon_class = None
pos_token_sentiment = None

negations = ['nicht', 'nein', 'es ist nicht', 'weder', 'noch', 'noch nie', 'kaum', 'selten', 'ist nicht',
       'sollte nicht', 'konnte nicht', 'wird nicht', 'kann nicht ', 'niemand', 'nichts', 'nichtigkeit']


import abc
from typing import Dict

class Token(metaclass=abc.ABCMeta):
    """
    Represents a generic **token**, which is a sequence of characters in the input.
    A token is identified using a tokenizer, which breaks up the sequence of characters in the input into a sequence
    of tokens.
    """

    @abc.abstractmethod
    def sentiment(self) -> Dict[str, float]:
        """
        Returns the token polarities. <br/>
        Each polarity value is between 0 and 1. <br/>
        _p_ in _[0,1]_

        Returns:
            Dict[str, float]: Polarities type as keys, sentiment values as values.

                    Example:
                    token.text="fuck"

                    >> token.sentiment()
                    {
                     'pos' : 0.0,
                     'neg' : 1.0,
                     'obj' : 0.0
                    }
        """

    @abc.abstractmethod
    def text(self) -> str:
        """Returns the text of the token."""

    @abc.abstractmethod
    def shape(self) -> str:
        """Returns a transformed token version to show orthographic features of the token.

        Alphabetic characters are replaced by x or X, and numeric characters are replaced by d, and sequences of the
        same character are truncated after length 4. For example, "Xxxx" or "dd".
        """

    @abc.abstractmethod
    def lower(self) -> str:
        """Returns the lowercase version of the token text."""

    @abc.abstractmethod
    def stem(self) -> str:
        """Returns the stem of the token text."""

    @abc.abstractmethod
    def lemma(self) -> str:
        """Returns the lemma of the token text."""

    @abc.abstractmethod
    def pos(self) -> str:
        """Returns the pos tag of the token text."""

    @abc.abstractmethod
    def dep(self) -> str:
        """Returns the dep tag of the token text."""

    @abc.abstractmethod
    def lang(self) -> str:
        """Returns the language of the token."""

    @abc.abstractmethod
    def is_stopword(self) -> bool:
        """Returns True if the token text is a stopword else False."""

    @abc.abstractmethod
    def is_ngram(self) -> bool:
        """Returns True if the token text is a ngram else False."""

    @abc.abstractmethod
    def is_digit(self) -> bool:
        """Returns True if the token text is a digit else False."""

    @abc.abstractmethod
    def is_upper(self) -> bool:
        """Returns True if the token text is upper else False."""

    @abc.abstractmethod
    def is_capitalize(self) -> bool:
        """Returns True if the token text is capitalize else False."""

    @abc.abstractmethod
    def is_punct(self) -> bool:
        """Returns True if the token text is a punctuation symbol else False."""

    @abc.abstractmethod
    def is_space(self) -> bool:
        """Returns True if the token text is a space character else False."""

    @abc.abstractmethod
    def is_bracket(self) -> bool:
        """Returns True if the token text is a bracket symbol else False."""

    @abc.abstractmethod
    def is_ascii(self) -> bool:
        """Returns True if the token text is encoded in ascii else False."""

    @abc.abstractmethod
    def is_negated(self) -> bool:
        """Returns True if the token text is negated in the sentence."""

    @abc.abstractmethod
    def add_metadata(self, metadata: dict) -> None:
        """Adds metadata as attributes of the token."""

    @abc.abstractmethod
    def to_dict(self) -> dict:
        """Converts the token object to a dict record."""



__all__ = ['vectorizer']
from compling.lexicalAnalysis.vectorization.impl.vectorizer import Vectorizer
from compling.lexicalAnalysis.vectorization.impl.nevectorizer import NEVectorizer
from compling.lexicalAnalysis.vectorization.impl.vsm import VSM


language = 'romanian'
iso639 = 'ro'

spacy_model_sm = iso639 + '_core_news_sm'
spacy_model_md = iso639 + '_core_news_md'
spacy_model_lg = iso639 + '_core_news_lg'

sentiment_lexicon_class = None
pos_token_sentiment = None

negations = ['nu', 'nici unul', 'nici', 'niciodată', 'nimeni', 'nimic', 'nonentitate', 'nulitate', 'neant']


import abc
import compling.nlp
from typing import *


class Tokenization(metaclass=abc.ABCMeta):
    """
    Tokenization converts input text to streams of tokens, where each token is a separate word, punctuation sign,
    number/amount, date, etc.
    """

    def __init__(self) -> None:
        """
        **\_\_init\_\_**: Creates a new Tokenization object.
        """

        # natural language processing object
        self.nlp = compling.nlp.NLP()

        self.text_key = self.nlp.config.get(s='Corpus', k='text_key')
        self.date_format = self.nlp.config.get(s='Corpus', k='date_format')

    @abc.abstractmethod
    def ngrams2tokens(self,
                      n: Union[int, Iterable[int]],
                      docs_in: Iterable[Dict[str, str]],
                      docs_out: Iterable[Dict[str, str]],
                      pos: Tuple[str, ...] = ("PROPN", "VERB", "NOUN", "ADJ"),
                      corpus_threshold: int = 50,
                      doc_threshold: int = 0,
                      len_gram: int = 3,
                      include: Tuple[str, ...] = None,
                      replace: bool = True) -> Iterable[Dict[int, str]]:
        """
        Calculates frequent n-grams in the corpus. They will be considered as unique tokens during the Tokenization task. <br/>
        Frequent n-grams will be calculated based on the documents in the input stream (docs_in). <br/>
        Frequent n-grams will be considered as unique tokens in the output stream documents (docs_out).

        Args:
           docs_in (Iterable[Dict[str, str]]): Stream of input json documents as dicts. <br/> Each document must have a key where the text of the document is stored. Frequent n-grams will be calculated by scrolling this stream.
           docs_out (Iterable[Dict[str, str]]): Stream of output json documents as dicts. <br/> Each document must have a key where the text of the document is stored. Frequent n-grams will be considered as unique tokens for each document in the stream. Each document will be yielded as output of this function.
           n (Union[int, Iterable[int]]): If int, the size of n-grams. (e.g. n=2, bigrams) <br/> If Iterable[int], the sizes of n-grams. (e.g. n=[2,3,5], bigrams, trigrams, fivegrams)
           pos (Tuple[str, ...], optional): default ('PROPN', 'VERB', 'NOUN', 'ADJ'). <br/> Part of speech of the first and the last token that make up n-grams. <br/> Filters the most informative n-grams. <br/> If None, the filter will be ignored.

                   Example:
                   pos = ("PROPN", "VERB", "NOUN", "ADJ")

                   These n-grams are IGNORED:
                   - "of a": of [ADP], a [DET]
                   - "at home":  at [ADP], home [NOUN]
                   - "as much then": as [ADP], as [ADP]
                   - "a computer scientist": a [DET], scientist [NOUN]
                   - "of natural phenomena": of [ADP], phenomena [NOUN]
                   - ...

                   These n-grams are CONSIDERED:
                   * "mother Earth": mother [NOUN], Earth [PROPN]
                   * "John likes": John [PROPN], likes [VERB]
                   * "computer scientist": computer [NOUN], scientist [NOUN]
                   * "Galilean scientific method": Galilean [ADJ], method [NOUN]
                   * "understanding of natural phenomena": understanding [NOUN], phenomena [NOUN]
                   ...

           corpus_threshold (int, optional, default=50): Filters n-grams that have a corpus frequency greater than corpus_threshold.
           doc_threshold (int, optional, default=0): Filters n-grams that have frequency greater than doc_threshold. <br/> The frequency of an n-gram in the corpus is the sum of the frequency of that n-gram in documents where the ngram occurs at least doc_thresold times.
           len_gram (int, optional, default=3): The size of the first and the last token that make up n-grams must be at least 'len_gram'.

                   Example:
                   len_gram = 5

                   These n-grams are IGNORED:
                   - "John likes": John [4], likes [5]
                   - "New York":  New [3], York [3]
                   - ...

                   These n-grams are CONSIDERED:
                   * "mother Earth": mother [6], Earth [5]
                   * "computer scientist": computer [8], scientist [9]
                   ...

           include (Tuple[str, ...], optional, default=None): Include a list of arbitrary n-grams.
           replace (bool, optional, default=True): If True, replaces a n-gram with its tokens separated by '_'. <br/> Else, concatenates a new token, made merging the n-gram tokens with '_', to the n-gram.

                   Example:

                    - replace = True:
                      "New York is the most populous city in the United States."
                       ->
                      "New_York is the most populous city in the United_States."

                    - replace = False:
                      "New York is the most populous city in the United States."
                       ->
                      "New York New_York is the most populous city in the United States United_States."

        Returns:
            Iterable[Dict[str, str]]: Stream of json documents (docs_out) with frequent n-grams considered as unique token.


            It will be setted _self.corpus_ngram_ (Dict[str, int]): N-grams as keys, frequencies as values. <br/> The frequencies of the n-grams included arbitrarily are set to 0.

                    Example:
                    include = ("Giorgio Napolitano")

                     {
                        "New York": 51,
                        "Giorgio Napolitano": 0,
                        ...
                        "Donald Trump": 402
                     }
        """

    @abc.abstractmethod
    def run(self,
            docs: Iterable[Dict[str, str]],
            doc_id: int = 0,
            token_id: int = 0,
            sent_id: int = 0,
            para_id: int = 0,
            para_size: int = 3,
            index_doc: bool = True,
            index_sent: bool = True,
            index_para: bool = True) -> Iterable[Dict[str, dict]]:
        """
        Runs the tokenization for each document of the corpus.

        Args:
            docs (Iterable[Dict[str, str]]): Stream of json documents as dicts. <br/> Each document must have a key where the text of the document is stored. Each document will be tokenized.
            doc_id (int, optional, default=0): Unique document identifier. <br/> Numeric id of documents. <br/> A progressive id will be assigned to the documents starting from this number.
            token_id (int, optional, default=0): Unique token identifier. <br/> Numeric id of tokens. <br/> A progressive id will be assigned to the tokens starting from this number.
            sent_id (int, optional, default=0): Unique sentence identifier. <br/> Numeric id of sentences. <br/> A progressive id will be assigned to the sentences starting from this number.
            para_id (int, optional, default=0): Unique paragraph identifier. <br/> Numeric id of paragraphs. <br/> A progressive id will be assigned to the paragraphs starting from this number.
            para_size (int, optional, default=3): The paragraph size: the number of sentences that makes up a paragraph.
            index_doc (bool, optional, default=True): If True, returns the records of the documents index as values of the key 'documents'.
            index_sent (bool, optional, default=True): If True, returns the records of the sentences index as values of the key 'sentences'.
            index_para (bool, optional, default=True): If True, returns the records of the paragraphs index as values of the key 'paragraphs'.

        Returns:
            Iterable[Dict[str, dict]]: Stream of tokens, sentences, paragraphs, documents records.

                    Example:
                    [index_sent == True, index_para == False, index_doc == True]

                    >>> t = Tokenizer(domain_stopwords=my_domain_stopwords)
                    >>> records = t.run(docs=my_docs):
                    >>> next(records).keys()
                    dict_keys(["tokens", "sentences", "documents"])


                    Example:
                    >>> record = next(records)
                    >>> tokens = record["tokens"]
                    >>> sentences = record["sentences"]
                    >>> print(tokens)
                    [
                        {
                        "token_id_": 0,
                        "sent_id_": 0,
                        "para_id_": 0,
                        "doc_id_": 0,
                        "text_": "When",
                        "sent_pos_": 0,
                        "para_pos_": 0,
                        "doc_pos_": 0,
                        "lower_": "when",
                        ...
                        },
                        ...
                    ]
                    >>> print(sentences)
                    [
                        {
                         "sent_id_": 0,
                         "text_": "When We , though all unworthy , were called to succeed on the Apostolic Throne the meek Pius X , whose life of holiness and well doing was cut short by grief -atti_degli_apostoli- the fratricidal struggle that had just burst forth in Europe , We , too , on turning a fearful glance on the blood stained battlefields , felt the anguish of a father , who sees his homestead devastated and in ruins before the fury of the hurricane .",
                         "named_entities_": ["Apostolic Throne", "Pius X", "Europe"],
                        ...
                        }
                    ]


                Check the config.ini file for records information.
        """



import abc
import compling.lexicalAnalysis.vectorization

class UnsupervisedLearning(compling.lexicalAnalysis.vectorization.VSM, metaclass=abc.ABCMeta):
    """
    Unsupervised learning is a type of machine learning that looks for previously undetected patterns in a data set with no pre-existing labels and with a minimum of human supervision.
    """

    @abc.abstractmethod
    def run(self, *args) -> object:
        """Runs Unsupervised Learning stage."""



from vaderSentiment.vaderSentiment import SentimentIntensityAnalyzer
from compling.analysis.sentiment.lexicon.lexicon import Lexicon


class Vader(Lexicon):
    """VADER Lexicon Implementation of Lexicon Abastract Class."""

    def __init__(self) -> None:
        """**\_\_init\_\_**: Creates a new Vader object."""
        self.vader = SentimentIntensityAnalyzer()

    def polarity(self, token: str) -> dict:
        """
        Returns the token polarities.

        Args:
            token (str): The text of the input token.

        Args:
            dict: Polarities type as keys, sentiment values as values.

                    Example:
                    {
                     'pos' : 0.9,
                     'neg' : 0.1,
                     'obj' : 0.3
                    }
        """

        polarities = self.vader.polarity_scores(token.replace("NOT_", ""))

        # polarity inversion
        if token.startswith('NOT_'):
            polarities['neg'], polarities['pos'] = polarities['pos'], polarities['neg']

        polarities['obj'] = polarities['neu']
        del polarities['neu']
        del polarities['compound']

        return polarities


import numpy as np
import pandas as pd
from typing import *
import matplotlib.pyplot as plt
from sklearn.metrics.pairwise import pairwise_distances


class VSM:
    """
    _Vector Space Model_ is an algebraic model for representing text documents as vectors of index terms.

    A VSM object allow you to easily explore the multi dimensional Vector Space.

    It is required that you have already performed the Vectorization.
    """

    def __init__(self, vectors: Iterable[dict], grouped_by_field: str, date_field=None) -> None:
        """
        **\_\_init\_\_**: Creates a new VSM object.

        Args:
            vectors (Iterable[dict]): Vector records produced by vectorization. <br/> Each record must have a "bow" (Bag of Words) field in which is stored a Dict[str, int]. It has tokens as key and frequencies as values.
            grouped_by_field (Union[str, List[str]]): The fields which the tokens have been grouped by (e.g. doc_id, sent_id, metada1, ...).
            date_field (str, optional, default=None): The field containing a date that refers to vectors (e.g. could be a 'publication_date' metadata). <br/> If not None, the data will be sorted by date_value.
        """

        if isinstance(grouped_by_field, str):
            self.grouped_by_field = [grouped_by_field]
        else:
            self.grouped_by_field = grouped_by_field

        self.grouped_by_field_key = " + ".join([g+'_' for g in self.grouped_by_field])
        self.date_field = date_field if date_field is None else date_field + '_'

        # dbmanager for transparent database management
        self.vectors = vectors

        # DataFrame conteining data
        self.data = {}
        for record in vectors:
            if date_field:
                self.data[record[self.grouped_by_field_key]] = dict(record['bow'], id_date=record[self.date_field])
            else:
                self.data[record[self.grouped_by_field_key]] = dict(record['bow'])
        self.data = pd.DataFrame.from_dict(self.data)

        # Replace nan with 0
        self.data.fillna(0, inplace=True)

        if date_field:
            # Sort by date_field
            self.data = self.data.T.sort_values('id_date', ascending=False)
            self.data = self.data.sort_values('id_date', axis=0)
            del self.data['id_date']

        self.distance_ = None

    def distance(self, metric: Union[str, Callable[[np.ndarray, np.ndarray], float], np.array] = "cosine") -> object:
        """
        Calculates the vector distance matrix between vectors.

        Args:

            metric (Union[str, callable], default==cosine): You can choose a metric from from sklearn.metrics.pairwise import pairwise_distances. <br/> See https://scikit-learn.org/stable/modules/generated/sklearn.metrics.pairwise_distances.html <br/><br/> You can also pass a function too.

                       def metric(np.array, np.array) -> np.array

        Returns:
            DataFrame: Distance matrix.
        """
        if type(metric) is str:
            metric = pairwise_distances(self.data.T, self.data.T, metric=metric)
        else:
            metric = metric(self.data.T, self.data.T)

        self.distance_ = pd.DataFrame(metric, index=self.data.T.index, columns=self.data.T.index)
        return self.distance_

    def plot(self, figsize: Tuple[int, int] = (10, 10), title: str = None, save:str=None) -> None:
        """
        Plot the distance matrix as a hitmap.

        Args:
            figsize (Tuple[int, int], optional, default=10, 10): Figure size: width, height in inches.
            save (str, optional, default=None): Figure is not saved if save is None. <br/> Set it with an output name if you want save it.
            title (str, optional, default=None):  If not None, draws title string on the figure.
        """

        if not hasattr(self, 'distance_'):
            self.distance()

        fig, ax = plt.subplots(figsize=figsize)
        ax.imshow(self.distance_)

        ax.set_xticks(np.arange(self.distance_.shape[0]))
        ax.set_yticks(np.arange(self.distance_.shape[0]))
        ax.set_xticklabels(self.distance_.index, rotation=90)
        ax.set_yticklabels(self.distance_.index)

        if title is not None:
            plt.title(title)

        if save:
            plt.tight_layout()
            fig1 = plt.gcf()
            plt.show()
            plt.draw()
            fig1.savefig(save)
        else:
            plt.show()

        plt.close()

    def topn(self, n: int = 10) -> Dict[str, Dict[str, float]]:
        """
        Returns the top N frequency values for each different self.grouped_by_field values.

        Args:
            n (int, optional, default=10): It will be returned the top 'n' values.

        Returns:
            Dict[str, Dict[str, float]]: grouped_by_field values as keys, frequency dict as values.

                    Example:
                    gropued_by_field = ['para_id', 'author']; n = 10

                    {
                     "para_1 + author1": {
                                         token1: 1.2,
                                         ...,
                                         token10: 0.01
                                        },
                     ...,
                     "para_151 + author1": {
                                         token1: 1.6,
                                         ...,
                                         token10: 0.5
                                        },
                     ...,
                     }
        """

        data = self.data.to_dict()

        topn = {}
        for key, value in data.items():
            topn[key] = dict(sorted(value.items(), key=lambda x: x[1], reverse=True)[:n])

        return topn